/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
#include "data.h"
#include "obj_any.h"
#include "conf_core.h"

#include "draw.h"

/* Draws a single object; returns whether the object needs overlay drawing */
RND_INLINE int osm_draw_obj(osm_layer_t *ly, osm_any_obj_t *obj, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform_caller, int overlay)
{
	obj->hdr.calls->draw(obj, gc, region, xform_caller, overlay);
	if (conf_core.appearance.bbox_debug) {
		rnd_coord_t x1, y1, x2, y2;

		x1 = obj->hdr.bbox.x1; y1 = obj->hdr.bbox.y1;
		x2 = obj->hdr.bbox.x2; y2 = obj->hdr.bbox.y2;

		rnd_hid_set_line_cap(gc, rnd_cap_round);
		rnd_hid_set_line_width(gc, -1);
		rnd_render->draw_line(gc, x1, y1, x2, y1);
		rnd_render->draw_line(gc, x2, y1, x2, y2);
		rnd_render->draw_line(gc, x2, y2, x1, y2);
		rnd_render->draw_line(gc, x1, y2, x1, y1);
	}

	TODO("return overlay draw request here");
	return 0;
}

void osm_draw_layer(osm_layer_t *ly, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform_caller)
{
	osm_rtree_it_t it;
	void *o;
	osm_rtree_box_t *view = (osm_rtree_box_t *)&region->view;
	int z;

	if (ly->plugin_draw != NULL) {
		ly->plugin_draw(ly, gc, region, xform_caller);
		return;
	}

	for(z = 0; z < OSM_ZORD_MAX; z++) {
		int need_overlay = 0;

		for(o = osm_rtree_first(&it, &ly->zord[z], view); o != NULL; o = osm_rtree_next(&it))
			need_overlay |= osm_draw_obj(ly, o, gc, region, xform_caller, 0);

		if (need_overlay)
			for(o = osm_rtree_first(&it, &ly->zord[z], view); o != NULL; o = osm_rtree_next(&it))
				need_overlay |= osm_draw_obj(ly, o, gc, region, xform_caller, 1);
	}
}

rnd_cardinal_t osm_draw_inhibit = 0;

void osm_expose_main(rnd_hid_t *hid, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform_caller)
{
	rnd_cardinal_t lid;
	rnd_hid_gc_t gc;
	static const int direct = 1;
	rnd_hid_t *save = rnd_render;
	rnd_xform_t xform_local;

	if (osm_draw_inhibit)
		return;

	if (xform_caller == NULL)
		xform_caller = &xform_local;

	xform_caller->zoom_mpp = (double)region->coord_per_pix / 100.0;

	if ((rnd_render != NULL) && (!rnd_render->override_render))
		rnd_render = hid;

	gc = rnd_render->make_gc(rnd_render);

	/* announce start of rendering */
	rnd_render->render_burst(rnd_render, RND_HID_BURST_START, &region->view);

	/* the HID needs a group set, even if we don't do groups */
	if (rnd_render->set_layer_group != NULL)
		rnd_render->set_layer_group(rnd_render, &osm.hidlib, 0, NULL, 0, 0, 0, 0, NULL);

	for(lid = 0; lid < osm.layers.used; lid++) {
		osm_layer_t *ly = osm.layers.array[lid];

		if (!ly->vis) continue;

		rnd_render->set_drawing_mode(rnd_render, RND_HID_COMP_RESET, direct, &region->view);
		rnd_render->set_drawing_mode(rnd_render, RND_HID_COMP_POSITIVE, direct, &region->view);
		osm_draw_layer(ly, gc, region, xform_caller);
		rnd_render->set_drawing_mode(rnd_render, RND_HID_COMP_FLUSH, direct, &region->view);
		rnd_render->end_layer(rnd_render);
	}

	rnd_render->render_burst(rnd_render, RND_HID_BURST_END, &region->view);
	rnd_render->destroy_gc(gc);

	rnd_render = save;
}

void osm_expose_preview(rnd_hid_t *hid, rnd_hid_expose_ctx_t *e)
{

}
