/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <ctype.h>
#include <librnd/config.h>
#include <librnd/core/actions.h>
#include <librnd/hid/hid.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/funchash_core.h>

#include "build_run.h"
#include "data.h"
#include "search.h"
#include "obj_node.h"
#include "funchash_core.h"

static const char osm_acts_Quit[] = "Quit()";
static const char osm_acth_Quit[] = "Quits the application.";
static fgw_error_t osm_act_Quit(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_rnd_quit_app();
	return FGW_SUCCESS;
}

extern fgw_error_t rnd_gui_act_zoom(fgw_arg_t *res, int argc, fgw_arg_t *argv);

TODO("Figure how to avoid copying this from librnd rnd_gui_acts_zoom")
const char osm_acts_Zoom[] = \
	"Zoom()\n" \
	"Zoom([+|-|=]factor)\n" \
	"Zoom(x1, y1, x2, y2)\n" \
	"Zoom(?)\n" \
	"Zoom(get)\n" \
	"Zoom(obj, type, id)\n" \
	"Zoom(obj, type #id)\n" \
	"Zoom(auto_first)\n";
const char osm_acth_Zoom[] = "GUI zoom";
fgw_error_t osm_act_Zoom(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	if ((argv[1].type & FGW_STR) == FGW_STR) {
		if (strcmp(argv[1].val.str, "auto_first") == 0) { /* zoom to extent when the first file is loaded (called from the menu file) */
			if (osm.layers.used == 1)
				return rnd_actionv_bin(RND_ACT_DESIGN, "rnd_zoom", res, 0, argv);
			return 0;
		}
		if (strcmp(argv[1].val.str, "obj") == 0) {
			osm_layer_t *ly = osm_layer_get_curr(&osm);
			char *ts, *is = NULL, *end;
			osm_objtype_t type;
			osm_id_t id;
			osm_any_obj_t *obj;
			fgw_arg_t args[5] = {0};
			rnd_coord_t mx, my;

			if (ly == NULL) {
				rnd_message(RND_MSG_ERROR, "Zoom obj: no layer selected");
				return -1;
			}

			RND_ACT_CONVARG(2, FGW_STR, Zoom, ts = argv[2].val.str);
			RND_ACT_MAY_CONVARG(3, FGW_STR, Zoom, is = argv[3].val.str);

			type = osm_str_to_obj_type(ts, &end);
			if (type == OSM_OBJ_invalid) {
				rnd_message(RND_MSG_ERROR, "Zoom obj: invalid type");
				return -1;
			}

			if (is == NULL) {
				is = end;
				while(isspace(*is)) is++;
				if (*is == '#') is++;
			}
			id = osm_str2id(is, &end);
			while(isspace(*end)) end++;
			if (*end != '\0') {
				rnd_message(RND_MSG_ERROR, "Zoom obj: invalid id");
				return -1;
			}

			obj = osm_obj_by_id(ly, type, id);
			if (obj == NULL) {
				rnd_message(RND_MSG_ERROR, "Zoom obj: object not found");
				return -1;
			}

			mx = (obj->hdr.bbox.x2 - obj->hdr.bbox.x1) / 4;
			my = (obj->hdr.bbox.y2 - obj->hdr.bbox.y1) / 4;
			if (mx < 100)
				mx = 5000;
			else if (mx > 100000)
				mx = 100000;
			if (my < 100)
				my = 5000;
			else if (my > 100000)
				my = 100000;
	
			args[1].type = FGW_COORD; fgw_coord(&args[1]) = obj->hdr.bbox.x1-mx;
			args[2].type = FGW_COORD; fgw_coord(&args[2]) = obj->hdr.bbox.y1-my;
			args[3].type = FGW_COORD; fgw_coord(&args[3]) = obj->hdr.bbox.x2+mx;
			args[4].type = FGW_COORD; fgw_coord(&args[4]) = obj->hdr.bbox.y2+my;
			return rnd_actionv_bin(RND_ACT_DESIGN, "rnd_zoom", res, 5, args);
		}
	}
	return rnd_actionv_bin(RND_ACT_DESIGN, "rnd_zoom", res, argc, argv);
}

static const char osm_acts_RefAng[] = "RefAng(set|deactivate)\nRefAng(activate, [1..8])";
static const char osm_acth_RefAng[] = "RefAngs the application.";
static fgw_error_t osm_act_RefAng(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	int op, idx;

	RND_ACT_CONVARG(1, FGW_KEYWORD, RefAng, op = fgw_keyword(&argv[1]));

	switch(op) {
		case F_Set:
			{
				vtp0_t lst = {0};
				osm_node_t *nd[2] = {0, 0};
				long n;
				osm_layer_t *ly = osm_layer_get_curr(osm);

				osm_search_all_selected(&lst, ly);
				for(n = 0; n < lst.used; n++) {
					osm_node_t *o = lst.array[n];
					if (o->hdr.type == OSM_OBJ_NODE) {
						if (nd[0] == NULL) {
							nd[0] = o;
						}
						else if (nd[1] == NULL) {
							nd[1] = o;
							break;
						}
					}
				}
				vtp0_uninit(&lst);

				if (nd[1] != NULL) {
					rnd_coord_t dx, dy;

					if (osm->refang == NULL)
						osm->refang = &osm->refangs[1];

					dx = nd[1]->x - nd[0]->x;
					dy = nd[1]->y - nd[0]->y;
					*osm->refang = atan2(dy, dx);
				}
			}
			break;

		case F_Deactivate:
			osm->refang = NULL;
			break;

		case F_Activate:
			RND_ACT_CONVARG(2, FGW_INT, RefAng, idx = argv[2].val.nat_int);
			if ((idx >= 0) && (idx < OSM_REFANG_MAX))
				osm->refang = &osm->refangs[idx];
			break;
	}

	return FGW_SUCCESS;
}


static rnd_action_t gui_action_list[] = {
	{"Quit", osm_act_Quit, osm_acth_Quit, osm_acts_Quit},
	{"Zoom", osm_act_Zoom, osm_acth_Zoom, osm_acts_Zoom},
	{"RefAng", osm_act_RefAng, osm_acth_RefAng, osm_acts_RefAng}
};

void gui_act_init(void)
{
	RND_REGISTER_ACTIONS(gui_action_list, NULL)
}

