/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <genvector/vtp0.h>
#include <genht/htpp.h>
#include <genht/hash.h>
#include <librnd/core/error.h>

#include "history.h"
#include "htIp.h"
#include "obj_any.h"

osm_history_t osm_hist = {0};
int osm_hist_inited = 0;

void osm_history_init(osm_history_t *hst)
{
	int t;
	for(t = 0; t < OSM_OBJ_vmax; t++)
		htIp_init(&hst->id2obj[t], htIp_keyhash, htIp_keyeq);
}

void osm_history_uninit(osm_history_t *hst)
{
	int t;

	for(t = 0; t < OSM_OBJ_vmax; t++) {
		htIp_entry_t *e;
		long n;

		for(e = htIp_first(&hst->id2obj[t]); e != NULL; e = htIp_next(&hst->id2obj[t], e)) {
			vtp0_t *lst = e->value;
			for(n = 0; n < lst->used; n++)
				osm_obj_free(lst->array[n]);
			vtp0_uninit(lst);
			free(lst);
		}
		htIp_uninit(&hst->id2obj[t]);
	}
}

void osm_history_add(osm_history_t *hst, osm_any_obj_t *obj)
{
	vtp0_t *lst = htIp_get(&hst->id2obj[obj->hdr.type], obj->hdr.id);
	long n, idx = -1;
	osm_any_obj_t *prev;

	if (lst == NULL) {
		lst = calloc(sizeof(vtp0_t), 1);
		htIp_set(&hst->id2obj[obj->hdr.type], obj->hdr.id, lst);
	}

	for(n = 0; n < lst->used; n++) {
		prev = lst->array[n];
		if (prev->hdr.version <= obj->hdr.version)
			idx = n;
	}

	/* new, earliest item, prepend */
	if (idx < 0) {
		vtp0_insert_len(lst, 0, (void **)&obj, 1);
		return;
	}

	/* same version: replace */
	prev = lst->array[idx];
	if (prev->hdr.version == obj->hdr.version) {
		osm_obj_free(prev);
		lst->array[idx] = obj;
		return;
	}

	/* new, latest item, append */
	if (idx >= lst->used) {
		vtp0_append(lst, obj);
		return;
	}

	/* new, after idx */
	vtp0_insert_len(lst, idx+1, (void **)&obj, 1);
}



static void timeline_add(osm_history_t *hst, htpp_t *work, vtp0_t *obj_versions)
{
	long n, i;
	osm_any_obj_t *obj = obj_versions->array[0]; /* enough to add the first version of each object */

	htpp_set(work, (void *)obj, (void *)obj);

	/* can't call the standard referee list callback because there's no layer or rtree */
	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:
		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_invalid:
		case OSM_OBJ_vmax:
			break;

		case OSM_OBJ_WAY: /* add all nodes of all obj versions of this way to work */
			for(i = 0; i < obj_versions->used; i++) {
				osm_any_obj_t *o = obj_versions->array[i];
				for(n = 0; n < o->way.curr->nd.used; n++) {
					vtp0_t *lst = htIp_get(&hst->id2obj[OSM_OBJ_NODE], o->way.curr->nd.array[n]);
					if ((lst != NULL) && (lst->used > 0))
						timeline_add(hst, work, lst);
				}
			}
			break;

		case OSM_OBJ_RELATION:
			TODO("relation: implement me");
			break;
	}
}


static osm_timeline_t *timeline_alloc(osm_changeset_t *chs, const osm_any_obj_t *subject)
{
	osm_timeline_t *t = calloc(sizeof(osm_timeline_t), 1);
	t->changeset = chs;
	t->subject = subject;
	
	return t;
}

static void timeline_insert_changeset(gdl_list_t *dst, osm_changeset_t *chs, const osm_any_obj_t *subject, osm_any_obj_t *chobj)
{
	osm_timeline_t *t, *prev = NULL;

	for(t = gdl_first(dst); t != NULL; t = t->link.next) {
		if (t->changeset->hdr.id < chs->hdr.id)
			prev = t;
		if (t->changeset->hdr.id == chs->hdr.id) {
			/* match - changeset already added */
			vtp0_append(&t->chobjs, chobj);
			return;
		}
	}

	/* changeset not already added, append after prev */
	t = timeline_alloc(chs, subject);
	vtp0_append(&t->chobjs, chobj);

	if (prev == NULL)
		gdl_insert(dst, t, link);
	else
		gdl_insert_after(dst, prev, t, link);
}

void osm_history_insert_timeline_for_obj(osm_history_t *hst, gdl_list_t *dst, osm_any_obj_t *obj)
{
	htpp_t work;
	htpp_entry_t *e;

	htpp_init(&work, ptrhash, ptrkeyeq);

	{ /* add the first version of obj from history (we hash by history obj ptrs) */
		vtp0_t *lst = htIp_get(&hst->id2obj[obj->hdr.type], obj->hdr.id);
		if ((lst != NULL) && (lst->used > 0))
			timeline_add(hst, &work, lst);
	}

	for(e = htpp_first(&work); e != NULL; e = htpp_next(&work, e)) {
		const osm_any_obj_t *o = e->key;
		vtp0_t *lst = htIp_get(&hst->id2obj[o->hdr.type], o->hdr.id);
		long n;

		for(n = 0; n < lst->used; n++) {
			osm_any_obj_t *ho = lst->array[n]; /* historic object */
			osm_id_t chs_id = ho->hdr.changeset;
			vtp0_t *chlst = htIp_get(&hst->id2obj[OSM_OBJ_CHANGESET], chs_id);
			osm_changeset_t *chs = NULL;

			if ((chlst != NULL) && (chlst->used > 0))
				chs = chlst->array[0];
			if (chs != NULL)
				timeline_insert_changeset(dst, chs, obj, ho);
		}
	}


	htpp_uninit(&work);
}

void osm_timeline_free(gdl_list_t *tl)
{
	osm_timeline_t *t;

	for(t = gdl_first(tl); t != NULL; t = gdl_first(tl)) {
		gdl_remove(tl, t, link);
		vtp0_uninit(&t->chobjs);
		free(t);
	}
}


void osm_history_draw_reset(osm_design_t *osm)
{
	osm_layer_t *ly = osm_layer_by_name(osm, "History", 0);
	if (ly != NULL)
		osm_layer_clear_objs(ly);
}

static void history_draw(osm_history_t *hst, osm_layer_t *ly, osm_any_obj_t *obj)
{
	long n, i;

	/* avoid adding duplicates */
	if (htIp_get(&ly->id2obj[obj->hdr.type], obj->hdr.id) != NULL)
		return;


	/* first draw all referees of the same age */
	switch(obj->hdr.type) {
		case OSM_OBJ_WAY:
			for(n = 0; n < obj->way.curr->nd.used; n++) {
				vtp0_t *lst = htIp_get(&hst->id2obj[OSM_OBJ_NODE], obj->way.curr->nd.array[n]);
				osm_any_obj_t *o, *best = NULL;

				if (lst != NULL) {
					for(i = 0; i < lst->used; i++) {
						o = lst->array[i];
						if (o->hdr.changeset <= obj->hdr.changeset)
							best = o;
					}
				}
				else
					rnd_message(RND_MSG_ERROR, "Can't find node #" OSM_PID " in the history\n", obj->way.curr->nd.array[n]);
				if (best != NULL)
					history_draw(hst, ly, best);
			}
			break;

		case OSM_OBJ_RELATION:
			TODO("relation: implement this");
			break;

		case OSM_OBJ_NODE: break;
		case OSM_OBJ_CHANGESET: break;
		case OSM_OBJ_vmax: break;
		case OSM_OBJ_invalid: break;
	}

	/* then draw the actual object */
	osm_obj_dup_into(obj, ly, obj->hdr.id);
}

void osm_history_draw_obj(osm_design_t *osm, osm_history_t *hst, osm_any_obj_t *obj)
{
	osm_layer_t *ly = osm_layer_by_name(osm, "History", 1);
	ly->editable = 0;

	history_draw(hst, ly, obj);
}

void osm_history_draw_objs(osm_design_t *osm, osm_history_t *hst, vtp0_t *objs)
{
	long n;
	osm_layer_t *ly = osm_layer_by_name(osm, "History", 1);
	ly->editable = 0;

	for(n = 0; n < objs->used; n++)
		history_draw(hst, ly, objs->array[n]);
}

