/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/config.h>
#include <librnd/core/compat_misc.h>

#include <ctype.h>
#include "style.h"
#include "obj_any.h"
#include "obj_node.h"
#include "obj_way.h"
#include "obj_relation.h"
#include "obj_common.h"

const char *osm_obj_name[] = {
	"node",
	"way",
	"relation",
	"changeset",
	/* WHEN UPDATING: don't forget to also update osm_str_to_obj_type() */
	NULL 
};

osm_objtype_t osm_str_to_obj_type(const char *str, char **end)
{
	while(isspace(*str)) str++;
	switch(*str) {
		case 'n': case 'N':
			if (rnd_strncasecmp(str, "node", 4) == 0) {
				if (end != NULL)
					*end = (char *)str + 4;
				return OSM_OBJ_NODE;
			}
			break;
		case 'w': case 'W':
			if (rnd_strncasecmp(str, "way", 3) == 0) {
				if (end != NULL)
					*end = (char *)str + 3;
				return OSM_OBJ_WAY;
			}
			break;
		case 'r': case 'R':
			if (rnd_strncasecmp(str, "relation", 8) == 0) {
				if (end != NULL)
					*end = (char *)str + 8;
				return OSM_OBJ_RELATION;
			}
			break;
		case 'c': case 'C':
			if (rnd_strncasecmp(str, "changeset", 9) == 0) {
				if (end != NULL)
					*end = (char *)str + 9;
				return OSM_OBJ_CHANGESET;
			}
			break;
	}

	*end = (char *)str;
	return OSM_OBJ_invalid;
}

osm_objtype_t osm_str_to_obj_type_id(const char *str, char **end, osm_id_t *id_out)
{
	osm_objtype_t ty;
	char *end_;

	if (end == NULL)
		end = &end_;

	ty = osm_str_to_obj_type(str, end);
	if (ty == OSM_OBJ_invalid)
		return OSM_OBJ_invalid;

	while(isspace(**end)) (*end)++;
	if (**end == ':') (*end)++;
	if (**end == '#') (*end)++;

	*id_out = osm_str2id(*end, end);
	if ((**end != '\0') && (!isspace(**end)))
		return OSM_OBJ_invalid;
	if (*id_out == 0)
		return OSM_OBJ_invalid;

	return ty;
}


void osm_obj_add_to_layer(osm_layer_t *ly, osm_any_obj_t *obj)
{
	assert(obj->hdr.type > OSM_OBJ_invalid);
	assert(obj->hdr.type < OSM_OBJ_max);
	assert(obj->hdr.layer == NULL);

	/* register in the layer */
	obj->hdr.layer = ly;
	osm_rtree_insert(&ly->objs, obj, osm_obj_bbox(obj));
	obj->relation.hdr.in_objs = 1;
	htIp_set(&ly->id2obj[obj->hdr.type], obj->hdr.id, obj);

	/* figure style and add to zord rtrees */
	osm_style_update_obj(obj);

	/* get caches recomputed */
	ly->rev++;
}

void osm_obj_del_from_layer(osm_any_obj_t *obj)
{
	osm_layer_t *ly = obj->hdr.layer;

	assert(obj->hdr.type > OSM_OBJ_invalid);
	assert(obj->hdr.type < OSM_OBJ_max);
	assert(obj->hdr.layer != NULL);

	osm_objcom_rtree_unreg(obj);
	obj->hdr.layer = NULL;
	htIp_pop(&ly->id2obj[obj->hdr.type], obj->hdr.id);

	/* get caches recomputed */
	ly->rev++;
}


long osm_obj_list_referers(vtp0_t *dst, osm_any_obj_t *obj, osm_objtype_mask_t mask)
{
	long count = 0;

	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:
			if (mask & OSM_MASK_WAY)
				count += osm_way_list_referers_by_node(dst, &obj->node);
			if (mask & OSM_MASK_RELATION)
				count += osm_relation_list_referers_by_obj(dst, obj);
			break;
		case OSM_OBJ_WAY:
			if (mask & OSM_MASK_RELATION)
				count += osm_relation_list_referers_by_obj(dst, obj);
			break;
		case OSM_OBJ_RELATION:
			if (mask & OSM_MASK_RELATION)
				count += osm_relation_list_referers_by_obj(dst, obj);
			break;

		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_vmax:
		case OSM_OBJ_invalid:
			break;
	}

	return count;
}

long osm_obj_dirt_referers(osm_any_obj_t *obj, osm_objtype_mask_t mask)
{
	long count = 0;

	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:
			if (mask & OSM_MASK_WAY)
				count += osm_way_dirt_referers_by_node(&obj->node);
			if (mask & OSM_MASK_RELATION)
				count += osm_relation_dirt_referers_by_obj(obj);
			break;
		case OSM_OBJ_WAY:
			if (mask & OSM_MASK_RELATION)
				count += osm_relation_dirt_referers_by_obj(obj);
			break;
		case OSM_OBJ_RELATION:
			if (mask & OSM_MASK_RELATION)
				count += osm_relation_dirt_referers_by_obj(obj);
			break;


		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_vmax:
		case OSM_OBJ_invalid:
			break;
	}

	return count;
}

void osm_obj_list_clear_referer(vtp0_t *lst)
{
	long n;
	for(n = 0; n < lst->used; n++) {
		osm_any_obj_t *o = lst->array[n];
		o->hdr.referer = 0;
	}
}

osm_rtree_box_t *osm_obj_bbox_update(osm_any_obj_t *obj)
{
	osm_objcom_rtree_unreg(obj);

	/* force-recalc */
	obj->hdr.bbox_valid = 0;
	osm_obj_bbox(obj);

	osm_objcom_rtree_reg(obj);
	return &obj->hdr.bbox;
}


#include "obj_any_undo.c"
