/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

/* included from obj_any.c to split the long code */

#include <librnd/core/rnd_printf.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/misc_util.h>
#include <librnd/core/error.h>
#include "undo.h"
#include "obj_node.h"
#include "obj_way.h"
#include "obj_relation.h"

static const char obj_any_cookie[] = "any_obj";

/* undoable fork if obj needs forking */
static void maybe_fork(osm_any_obj_t *obj)
{
	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:
			if (obj->node.curr == &obj->node.orig)
				osm_node_modify_fork(obj->node.hdr.layer, &obj->node, 1);
			break;

		case OSM_OBJ_WAY:
			if (obj->way.curr == &obj->way.orig)
				osm_way_modify_fork(obj->way.hdr.layer, &obj->way, 1);
			break;

		case OSM_OBJ_RELATION:
			if (obj->relation.curr == &obj->relation.orig) {
				abort();
				TODO("implement this for relation");
/*				osm_relation_modify_fork(obj->relation.hdr.layer, &obj->relation, 1);*/
			}
			break;

		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_invalid:
		case OSM_OBJ_vmax:
			abort();
	}
}

typedef struct {
	osm_layer_t *ly;
	osm_id_t id;
	osm_objtype_t type;

	long remove_list_slot; /* for non-blank objects */
} undo_obj_creat_rem_t;

static int undo_obj_creat_rem_swap(void *udata)
{
	undo_obj_creat_rem_t *u = udata;
	osm_any_obj_t *obj = osm_obj_by_id(u->ly, u->type, u->id);

	if (obj == NULL) {
		if (u->remove_list_slot < 0) {
			switch(u->type) {
				case OSM_OBJ_NODE:      obj = (osm_any_obj_t *)osm_node_new(); break;
				case OSM_OBJ_WAY:       obj = (osm_any_obj_t *)osm_way_new(); break;
				case OSM_OBJ_RELATION:  obj = (osm_any_obj_t *)osm_relation_new(); break;
				case OSM_OBJ_CHANGESET:
				case OSM_OBJ_invalid:
				case OSM_OBJ_vmax:       abort();
			}
		}
		else {
			/* resurrect the object from the remove list so data is preserved */
			obj = u->ly->remove_list.array[u->remove_list_slot];
			u->ly->remove_list.array[u->remove_list_slot] = NULL;
		}
		obj->hdr.id = u->id;
		obj->hdr.visible = 1;
		osm_obj_add_to_layer(u->ly, obj);
		if (u->remove_list_slot >= 0)
			osm_objcom_rtree_reg(obj);
	}
	else {
		osm_obj_del_from_layer(obj);
		obj->hdr.visible = 0;

		if (u->remove_list_slot < 0)
			osm_obj_free(obj);
		else
			u->ly->remove_list.array[u->remove_list_slot] = obj; /* teleport the object to the remove list so data is preserved */
	}

	return 0;
}

static void undo_obj_creat_rem_print(void *udata, char *dst, size_t dst_len)
{
	undo_obj_creat_rem_t *u = udata;
	osm_any_obj_t *obj = htIp_get(&u->ly->id2obj[u->type], u->id);
	const char *isblank = (u->remove_list_slot < 0) ? "blank" : "";

	if (obj == NULL)
		rnd_snprintf(dst, dst_len, "create %s %s #" OSM_PID, isblank, osm_obj_name[u->type], u->id);
	else
		rnd_snprintf(dst, dst_len, "remove %s %s #" OSM_PID, isblank, osm_obj_name[u->type], u->id);
}


static const uundo_oper_t undo_obj_creat_rem = {
	obj_any_cookie,
	NULL,
	undo_obj_creat_rem_swap,
	undo_obj_creat_rem_swap,
	undo_obj_creat_rem_print
};


osm_any_obj_t *osm_obj_create(osm_layer_t *layer, osm_objtype_t type, int undoable)
{
	undo_obj_creat_rem_t utmp, *u = &utmp;

	assert(type >= 0);
	assert(type < OSM_OBJ_max);

	if (undoable) u = uundo_append(&layer->undo, &undo_obj_creat_rem, sizeof(undo_obj_creat_rem_t));

	u->ly = layer;
	u->id = --layer->last_new_id;
	u->type = type;
	u->remove_list_slot = -1;

	undo_obj_creat_rem_swap(u);
	if (undoable) osm_undo_inc_serial(layer);

	return htIp_get(&u->ly->id2obj[u->type], u->id);
}


RND_INLINE void osm_obj_pre_remove_all_referees(osm_any_obj_t *remo, int undoable)
{
	osm_layer_t *ly = remo->hdr.layer;
	osm_rtree_it_t it;
	void *o;

	for(o = osm_rtree_first(&it, &ly->objs, &remo->hdr.bbox); o != NULL; o = osm_rtree_next(&it)) {
		osm_any_obj_t *referer = o;

		if (referer->hdr.type == OSM_OBJ_NODE) continue; /* otimization: nodes don't refer to other objects */

		/* there's no more efficient way than to do a search on all members anyway */
		referer->hdr.calls->pre_remove_referee(referer, remo->hdr.type, remo->hdr.id, undoable);
	}
}


void osm_obj_remove(osm_any_obj_t *obj, int undoable)
{
	undo_obj_creat_rem_t utmp, *u = &utmp;
	osm_layer_t *layer = obj->hdr.layer;

	assert(layer != NULL);

	if (undoable)
		uundo_freeze_serial(&layer->undo); /* need to batch because potential referees */

	osm_obj_pre_remove_all_referees(obj, undoable);

	if (undoable) u = uundo_append(&layer->undo, &undo_obj_creat_rem, sizeof(undo_obj_creat_rem_t));

	u->ly = obj->hdr.layer;
	u->id = obj->hdr.id;
	u->type = obj->hdr.type;

	if (undoable) {
		u->remove_list_slot = layer->remove_list.used;
		vtp0_append(&layer->remove_list, NULL);
	}
	else
		u->remove_list_slot = -1;

	undo_obj_creat_rem_swap(u);

	if (u->remove_list_slot >= 0)
		obj->hdr.calls->post_remove(obj, undoable);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}



typedef struct {
	osm_layer_t *ly;
	osm_id_t id;
	osm_objtype_t type;

	char *key, *val;
} undo_obj_tag_t;

static int undo_obj_tag_swap(void *udata)
{
	undo_obj_tag_t *u = udata;
	osm_any_obj_t *obj = osm_obj_by_id(u->ly, u->type, u->id);
	osm_tags_t *tags;
	htsp_entry_t *e;

	if (obj == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_obj_tag_swap(): object not found\n");
		return -1;
	}
	tags = obj->hdr.calls->get_tags(obj);

	e = htsp_getentry(tags, u->key);
	rnd_swap(char *, e->value, u->val);

	return 0;
}

static void undo_obj_tag_print(void *udata, char *dst, size_t dst_len)
{
	undo_obj_tag_t *u = udata;

	rnd_snprintf(dst, dst_len, "tag modify %s #" OSM_PID " %s: '%s'",
		osm_obj_name[u->type], u->id, u->key, u->val);
}

static void undo_obj_tag_free(void *udata)
{
	undo_obj_tag_t *u = udata;
	free(u->key); u->key = NULL;
	free(u->val); u->val = NULL;
}

static const uundo_oper_t undo_obj_tag = {
	obj_any_cookie,
	undo_obj_tag_free,
	undo_obj_tag_swap,
	undo_obj_tag_swap,
	undo_obj_tag_print
};


void osm_obj_modify_tag(osm_layer_t *layer, osm_any_obj_t *obj, const char *key, const char *new_val, int undoable)
{
	undo_obj_tag_t utmp, *u = &utmp;

	if (undoable) {
		uundo_freeze_serial(&layer->undo);
		maybe_fork(obj);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_obj_tag, sizeof(undo_obj_tag_t));


	u->ly = layer;
	u->type = obj->hdr.type;
	u->id = obj->hdr.id;
	u->key = rnd_strdup(key);
	u->val = rnd_strdup(new_val);

	undo_obj_tag_swap(u);
	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}


typedef struct {
	osm_layer_t *ly;
	osm_id_t id;
	osm_objtype_t type;

	char *key;
	char *val_saved;
} undo_tag_creat_rem_t;

static int undo_tag_creat_rem_swap(void *udata)
{
	undo_tag_creat_rem_t *u = udata;
	osm_any_obj_t *obj = osm_obj_by_id(u->ly, u->type, u->id);
	osm_tags_t *tags;

	if (obj == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_tag_creat_rem_swap(): object not found\n");
		return -1;
	}

	tags = obj->hdr.calls->get_tags(obj);
	if (htsp_has(tags, u->key)) {
		htsp_entry_t *e = htsp_popentry(tags, u->key);
		u->val_saved = e->value;
		e->value = NULL;
		free(e->key);
	}
	else {
		htsp_set(tags, rnd_strdup(u->key), u->val_saved);
		u->val_saved = NULL;
	}

	return 0;
}

static void undo_tag_creat_rem_print(void *udata, char *dst, size_t dst_len)
{
	undo_tag_creat_rem_t *u = udata;
	osm_any_obj_t *obj = htIp_get(&u->ly->id2obj[u->type], u->id);
	osm_tags_t *tags;

	if (obj == NULL) {
		rnd_snprintf(dst, dst_len, "tag create/remove: object not found: %s #" OSM_PID, osm_obj_name[u->type], u->id);
		return;
	}

	tags = obj->hdr.calls->get_tags(obj);
	if (!htsp_has(tags, u->key))
		rnd_snprintf(dst, dst_len, "create %s #" OSM_PID " key %s", osm_obj_name[u->type], u->id, u->key);
	else
		rnd_snprintf(dst, dst_len, "remove %s #" OSM_PID " key %s", osm_obj_name[u->type], u->id, u->key);
}

static void undo_tag_creat_rem_free(void *udata)
{
	undo_tag_creat_rem_t *u = udata;
	free(u->key); u->key = NULL;
	free(u->val_saved); u->val_saved = NULL;
}

static const uundo_oper_t undo_tag_creat_rem = {
	obj_any_cookie,
	undo_tag_creat_rem_free,
	undo_tag_creat_rem_swap,
	undo_tag_creat_rem_swap,
	undo_tag_creat_rem_print
};


void osm_tag_create(osm_any_obj_t *obj, const char *key, const char *new_val, int undoable)
{
	undo_tag_creat_rem_t utmp, *u = &utmp;
	osm_layer_t *layer = obj->hdr.layer;
	osm_tags_t *tags = obj->hdr.calls->get_tags(obj);

	assert(!htsp_has(tags, key));

	if (undoable) {
		uundo_freeze_serial(&layer->undo);
		maybe_fork(obj);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_tag_creat_rem, sizeof(undo_tag_creat_rem_t));

	u->ly = layer;
	u->type = obj->hdr.type;
	u->id = obj->hdr.id;
	u->key = rnd_strdup(key);
	u->val_saved = rnd_strdup(new_val);

	undo_tag_creat_rem_swap(u);
	if (undoable) osm_undo_inc_serial(layer);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}



void osm_tag_remove(osm_any_obj_t *obj, const char *key, int undoable)
{
	undo_tag_creat_rem_t utmp, *u = &utmp;
	osm_layer_t *layer = obj->hdr.layer;
	osm_tags_t *tags = obj->hdr.calls->get_tags(obj);

	assert(htsp_has(tags, key));

	if (undoable) {
		uundo_freeze_serial(&layer->undo);
		maybe_fork(obj);
	}

	assert(layer != NULL);

	if (undoable) u = uundo_append(&layer->undo, &undo_tag_creat_rem, sizeof(undo_tag_creat_rem_t));

	u->ly = layer;
	u->type = obj->hdr.type;
	u->id = obj->hdr.id;
	u->key = rnd_strdup(key);
	u->val_saved = NULL;

	undo_tag_creat_rem_swap(u);
	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}
