/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

/* Inline functions for common code for low level object implementation;
   only obj_*.c and style.c should use this */

#ifndef OSM_OBJ_COMMON_H
#define OSM_OBJ_COMMON_H

#include "config.h"

#include <librnd/core/error.h>
#include <librnd/core/conf.h>

#include "obj_any.h"
#include "style.h"

/* remove only from style-related zord rtree */
RND_INLINE void osm_objcom_rtree_unreg_style(osm_any_obj_t *o)
{
	if (o->hdr.zord >= 0)
		osm_rtree_delete(&o->hdr.layer->zord[(int)o->hdr.zord], o, &o->hdr.bbox);
	o->hdr.zord = -1;
}

/* Remove the object from any layer rtree */
RND_INLINE void osm_objcom_rtree_unreg(osm_any_obj_t *o)
{
	if (o->relation.hdr.in_objs) {
		o->relation.hdr.in_objs = 0;
		osm_rtree_delete(&o->hdr.layer->objs, o, &o->hdr.bbox);
	}

	osm_objcom_rtree_unreg_style(o);
}

/* register only in style-related zord rtree */
RND_INLINE void osm_objcom_rtree_reg_style(osm_any_obj_t *o)
{
	if (o->hdr.style != NULL) {
		if (o->hdr.zord < 0) {
			o->hdr.zord = o->hdr.style->zord;
			if (o->hdr.zord > OSM_ZORD_MAX-1)
				o->hdr.zord = OSM_ZORD_MAX-1;
			if (o->hdr.zord >= 0)
				osm_rtree_insert(&o->hdr.layer->zord[(int)o->hdr.zord], o, &o->hdr.bbox);
		}
	}
}


/* Call this to register an object in the rtrees. Assumesbbox is already
   calculated and valid so call bbox update right before this call */
RND_INLINE void osm_objcom_rtree_reg(osm_any_obj_t *o)
{
	assert(o->relation.hdr.bbox_valid);

	if (!o->relation.hdr.in_objs) {
		osm_rtree_insert(&o->hdr.layer->objs, o, &o->hdr.bbox);
		o->relation.hdr.in_objs = 1;
	}
	osm_objcom_rtree_reg_style(o);
}

/* Initialize the header of a new object */
RND_INLINE void osm_objcom_init(osm_hdr_t *hdr, osm_objtype_t type, const osm_objcalls_t *calls)
{
	hdr->type = type;
	hdr->calls = calls;
	hdr->in_objs = 0;
	hdr->zord = -1;
	hdr->style = NULL;
	hdr->visible = 1;
}

RND_INLINE void osm_objcom_diffresult(osm_any_obj_t *obj, osm_id_t new_id, osm_id_t new_version)
{
	osm_layer_t *ly = obj->hdr.layer;

	/* replace verion */
	if (obj->hdr.version > new_version) {
		rnd_message(RND_MSG_ERROR,
			"osm_objcom_diffresult(): unexpected version downgrade from " OSM_PID " to " OSM_PID " in object %s:" OSM_PID "->" OSM_PID "\n",
			obj->hdr.version, new_version, osm_obj_name[obj->hdr.type], obj->hdr.id, new_id);
	}
	obj->hdr.version = new_version;

	/* replace ID, updating the id2obj hash table */
	htIp_pop(&ly->id2obj[obj->hdr.type], obj->hdr.id);
	obj->hdr.id = new_id;
	htIp_set(&ly->id2obj[obj->hdr.type], obj->hdr.id, obj);
}

RND_INLINE const char *osm_objcom_visual_type(osm_any_obj_t *o, const rnd_conflist_t *lst, htsp_t *tags)
{
	rnd_conf_listitem_t *item_li;
	const char *item_str;
	int idx;

	rnd_conf_loop_list_str(lst, item_li, item_str, idx) {
		const char *res = htsp_get(tags, item_str);
		if (res != NULL) {
			if (strcmp(res, "yes") == 0) /* e.g. building=yes; return "building" */
				return item_str;
			if (strcmp(res, "no") == 0)  /* e.g. building=no; skip */
				continue;
			return res;
		}
	}

	return NULL;
}

#endif
