/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/config.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>

#include "conf_core.h"
#include "obj_node.h"
#include "obj_common.h"
#include "projection.h"
#include "draw.h"

static void osm_node_free_fields(osm_any_obj_t *obj)
{
	osm_tags_uninit(&obj->node.curr->tags);
	if (obj->node.curr != &obj->node.orig) {
		free(obj->node.curr);
		obj->node.curr = &obj->node.orig;
	}
}

/* Low level draw: assumes gc is set up to style */
RND_INLINE void osm_node_draw_(osm_any_obj_t *obj, rnd_hid_gc_t gc, int s, int att)
{
	rnd_coord_t x = att ? obj->node.att_x : obj->node.x;
	rnd_coord_t y = att ? obj->node.att_y : obj->node.y;

	rnd_render->draw_line(gc, x-s, y-s, x+s, y-s);
	rnd_render->draw_line(gc, x+s, y-s, x+s, y+s);
	rnd_render->draw_line(gc, x+s, y+s, x-s, y+s);
	rnd_render->draw_line(gc, x-s, y+s, x-s, y-s);
}

static void osm_node_draw(osm_any_obj_t *obj, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform, int overlay)
{
	rnd_coord_t cpp = region->coord_per_pix, s;

	if (obj->hdr.style == NULL) {
		rnd_hid_set_line_cap(gc, rnd_cap_round);
		rnd_hid_set_line_width(gc, -1);
		rnd_render->set_color(gc, rnd_color_magenta);
		s = 2*cpp;
	}
	else {
		osm_style_data_t *sd;

		if (!obj->hdr.selected && (xform->zoom_mpp > obj->hdr.style->lod_invis_above))
			return;
		else if (obj->hdr.selected || (xform->zoom_mpp < obj->hdr.style->lod_high_below))
			sd = &obj->hdr.style->high;
		else
			sd = &obj->hdr.style->low;

		rnd_hid_set_line_cap(gc, rnd_cap_round);
		rnd_hid_set_line_width(gc, -1);
		rnd_render->set_color(gc, obj->hdr.selected ? &conf_core.appearance.color.selected : &sd->stroke.color);
		s = sd->stroke.width;
		if (s < 0)
			s = -s * cpp;
	}

	osm_node_draw_(obj, gc, s, 0);
}

static void osm_node_draw_attached(osm_any_obj_t *obj, rnd_hid_gc_t gc)
{
	osm_node_draw_(obj, gc, rnd_gui->coord_per_pix << 2, 1);
}

static void osm_node_bbox(osm_any_obj_t *obj)
{
	lldeg2rnd(obj->node.curr->lon, obj->node.curr->lat, &obj->node.x, &obj->node.y);

#if 0
	{ /* projection round trip testing */
		double tlon, tlat;
		rnd2lldeg(&tlon, &tlat, obj->node.x, obj->node.y);
		rnd_trace("LLRT: %.7f %.7f -> %ld %ld -> %.7f %.7f\n", obj->node.curr->lon, obj->node.curr->lat, obj->node.x, obj->node.y, tlon, tlat);
	}
#endif

	obj->node.hdr.bbox.x1 = obj->node.x;
	obj->node.hdr.bbox.x2 = obj->node.x + 1;
	obj->node.hdr.bbox.y1 = obj->node.y;
	obj->node.hdr.bbox.y2 = obj->node.y + 1;
	obj->node.hdr.bbox_valid = 1;
}

static osm_any_obj_t *osm_node_alloc(void) { return (osm_any_obj_t *)osm_node_new(); }

static int osm_node_isc_box(const osm_any_obj_t *o, const osm_rtree_box_t *box)
{
	if (o->node.x < box->x1) return 0;
	if (o->node.x > box->x2) return 0;
	if (o->node.y < box->y1) return 0;
	if (o->node.y > box->y2) return 0;
	return 1;
}

static osm_tags_t *osm_node_get_tags(osm_any_obj_t *o)
{
	return &o->node.curr->tags;
}

void osm_node_data_copy(osm_node_data_t *dst, const osm_node_data_t *src, int deep)
{
	dst->lat = src->lat;
	dst->lon = src->lon;
	if (deep)
		osm_tags_dup(&dst->tags, &src->tags);
}

static void osm_node_fork(osm_node_t *nd)
{
	assert(nd->curr == &nd->orig);
	nd->curr = malloc(sizeof(osm_node_data_t));
	osm_node_data_copy(nd->curr, &nd->orig, 1);
}

static void osm_node_unfork(osm_node_t *nd)
{
	assert(nd->curr != &nd->orig);
	osm_tags_uninit(&nd->curr->tags);
	free(nd->curr);
	nd->curr = &nd->orig;
}


void osm_node_pre_edit(osm_any_obj_t *o, int geo, int auto_fork)
{
	if (auto_fork && (o->node.curr == &o->node.orig))
		osm_node_fork(&o->node);

	if (geo) {
		osm_obj_dirt_referers(o, OSM_MASK_WAY | OSM_MASK_RELATION);
		osm_objcom_rtree_unreg(o);
		o->node.hdr.bbox_valid = 0;
	}
}


void osm_node_post_edit(osm_any_obj_t *o, int tags_changed)
{
	if (tags_changed)
		osm_style_update_obj(o);
	if (!o->node.hdr.bbox_valid) {
		osm_node_bbox(o);
		osm_objcom_rtree_reg(o);
	}
	o->hdr.layer->rev++;
}

void osm_node_referee_objs(osm_any_obj_t *o, vtp0_t *dst, int update_att)
{
	if (update_att) {
		o->node.att_x = o->node.x;
		o->node.att_y = o->node.y;
	}
	if ((dst != NULL) && (!o->hdr.referer)) {
		vtp0_append(dst, o);
		o->hdr.referer = 1;
	}
}

static long osm_node_referee_idx(osm_any_obj_t *parent, osm_any_obj_t *child)
{
	return -1;
}


static void osm_node_diffresult(osm_any_obj_t *o, osm_id_t new_id, osm_id_t new_version)
{
	osm_tags_uninit(&o->node.orig.tags);
	osm_node_data_copy(&o->node.orig, o->node.curr, 0);
	o->node.orig.tags = o->node.curr->tags;
	free(o->node.curr);
	o->node.curr = &o->node.orig;
	osm_objcom_diffresult(o, new_id, new_version);
}

static void osm_node_diffresult_referees(osm_any_obj_t *o, htdr_t *dr, int permit_del, int undoable)
{
	/* nothing to do since nodes can not refer to anything */
}

static void osm_node_pre_remove_referee(osm_any_obj_t *o, osm_objtype_t rem_ty, osm_id_t rem_id, int undoable)
{
	/* nothing to do since nodes can not refer to anything */
}

static void osm_node_post_remove(osm_any_obj_t *o, int undoable)
{
	/* nothing to do since nodes can not refer to anything */
}

osm_any_obj_t *osm_node_dup_into(osm_any_obj_t *o, osm_layer_t *dst, osm_id_t new_id)
{
	osm_any_obj_t *obj = osm_node_alloc();
	osm_node_data_copy(obj->node.curr, o->node.curr, 1);
	if (new_id != 0)
		obj->hdr.id = new_id;
	else
		obj->hdr.id = o->hdr.id;
	osm_obj_add_to_layer(dst, obj);
	return obj;
}

static const char *osm_node_visual_type(osm_any_obj_t *o)
{
	return osm_objcom_visual_type(o, &conf_core.tagging.node_visual_types, &o->node.curr->tags);
}


static const osm_objcalls_t osm_node_calls = {
	osm_node_alloc,
	osm_node_free_fields,
	osm_node_draw,
	osm_node_draw_attached,
	osm_node_bbox,
	osm_node_isc_box,
	osm_node_get_tags,
	osm_node_pre_edit,
	osm_node_post_edit,
	osm_node_referee_objs,
	osm_node_referee_idx,
	osm_node_diffresult,
	osm_node_diffresult_referees,
	osm_node_pre_remove_referee,
	osm_node_post_remove,
	osm_node_dup_into,
	osm_node_visual_type
};

void osm_node_init(osm_node_t *node)
{
	memset(node, 0, sizeof(osm_node_t));
	osm_objcom_init(&node->hdr, OSM_OBJ_NODE, &osm_node_calls);
	node->curr = &node->orig;
	osm_tags_init(&node->orig.tags);
}

osm_node_t *osm_node_new(void)
{
	osm_node_t *res = malloc(sizeof(osm_node_t));
	osm_node_init(res);
	return res;
}

void osm_node_move_to_att(osm_node_t *node, int undoable)
{
	double lon, lat;

	if ((node->x == node->att_x) && (node->y == node->att_y))
		return;

	rnd2lldeg(&lon, &lat, node->att_x, node->att_y);

	osm_node_modify(node->hdr.layer, node, &lon, &lat, undoable);
}

int osm_node_is_orphan(osm_node_t *node, osm_any_obj_t *ignore_referer)
{
	osm_layer_t *ly = node->hdr.layer;
	osm_rtree_it_t it;
	void *o;

	/* First the cheap test: if it has tags, it's not orphaned */
	if (node->curr->tags.used > 0)
		return 0;


	/* find all referers */
	for(o = osm_rtree_first(&it, &ly->objs, &node->hdr.bbox); o != NULL; o = osm_rtree_next(&it)) {
		osm_any_obj_t *referer = o;

		if (referer->hdr.type == OSM_OBJ_NODE) continue; /* otimization: nodes don't refer to other objects */
		if (referer == ignore_referer) continue;

		if (referer->hdr.calls->osm_way_referee_idx(referer, (osm_any_obj_t *)node) >= 0)
			return 0;
	}

	/* nobody wanted this node */
	return 1;
}

#include "obj_node_undo.c"
