/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

/* included from obj_node.c to split the long code */

#include "undo.h"

static const char obj_node_cookie[] = "node";

typedef struct {
	osm_id_t id;
	osm_layer_t *ly;
} undo_node_fork_t;

static int undo_node_fork_swap(void *udata)
{
	undo_node_fork_t *u = udata;
	osm_node_t *node = htIp_get(&u->ly->id2obj[OSM_OBJ_NODE], u->id);

	if (node == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_node_fork_swap(): node not found: " OSM_PID "\n", u->id);
		return -1;
	}

	if (node->curr == &node->orig)
		osm_node_fork(node);
	else
		osm_node_unfork(node);

	return 0;
}

static void undo_node_fork_print(void *udata, char *dst, size_t dst_len)
{
	undo_node_fork_t *u = udata;
	osm_node_t *node = htIp_get(&u->ly->id2obj[OSM_OBJ_NODE], u->id);

	if (node == NULL) {
		rnd_snprintf(dst, dst_len, "node fork: invalid id " OSM_PID, u->id);
		return;
	}

	rnd_snprintf(dst, dst_len, "node %s", (node->curr == &node->orig) ? "fork" : "unfork");
}


static const uundo_oper_t undo_node_fork = {
	obj_node_cookie,
	NULL,
	undo_node_fork_swap,
	undo_node_fork_swap,
	undo_node_fork_print
};


void osm_node_modify_fork(osm_layer_t *layer, osm_node_t *node, int undoable)
{
	undo_node_fork_t utmp, *u = &utmp;

	if (undoable) u = uundo_append(&layer->undo, &undo_node_fork, sizeof(undo_node_fork_t));

	u->ly = layer;
	u->id = node->hdr.id;

	undo_node_fork_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
}




typedef struct {
	osm_id_t id;
	osm_layer_t *ly;
	double lon, lat;
} undo_node_modify_t;

static int undo_node_modify_swap(void *udata)
{
	undo_node_modify_t *u = udata;
	osm_node_t *node = htIp_get(&u->ly->id2obj[OSM_OBJ_NODE], u->id);

	if (node == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_node_modify_swap(): node not found: " OSM_PID "\n", u->id);
		return -1;
	}

	TODO("Figure redraw");
/* osm_obj_redraw(node); */


	osm_node_pre_edit((osm_any_obj_t *)node, 1, 1);

	rnd_swap(double, u->lon, node->curr->lon);
	rnd_swap(double, u->lat, node->curr->lat);

	osm_node_post_edit((osm_any_obj_t *)node, 0);

	TODO("Figure layer changed flag");
	TODO("Figure redraw");
/* osm_obj_redraw(node); */

	return 0;
}

static void undo_node_modify_print(void *udata, char *dst, size_t dst_len)
{
	undo_node_modify_t *u = udata;
	osm_node_t *node = htIp_get(&u->ly->id2obj[OSM_OBJ_NODE], u->id);

	if (node == NULL) {
		rnd_snprintf(dst, dst_len, "node modify: invalid id " OSM_PID, u->id);
		return;
	}

	rnd_snprintf(dst, dst_len, "node modify (%.7f;%.7f<->%.7f;%.7f)",
		u->lat, u->lon, node->curr->lat, node->curr->lon);
}

static void undo_node_modify_free(void *udata)
{
/*	undo_node_modify_t *u = udata;*/
}

static const uundo_oper_t undo_node_modify = {
	obj_node_cookie,
	undo_node_modify_free,
	undo_node_modify_swap,
	undo_node_modify_swap,
	undo_node_modify_print
};


void osm_node_modify(osm_layer_t *layer, osm_node_t *node, double *lon, double *lat, int undoable)
{
	undo_node_modify_t utmp, *u = &utmp;

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (node->curr == &node->orig)
			osm_node_modify_fork(layer, node, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_node_modify, sizeof(undo_node_modify_t));

	u->ly = layer;
	u->id = node->hdr.id;
	u->lon = OSM_UNDO_MODIFY(node, lon);
	u->lat = OSM_UNDO_MODIFY(node, lat);

	undo_node_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_node_modify_free(u);


	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

