/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/config.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
#include <librnd/core/compat_misc.h>


#include "obj_common.h"
#include "obj_relation.h"

static void osm_relation_free_fields(osm_any_obj_t *obj)
{
	long n;

	vtp0_uninit(&obj->relation.mbrp);
	obj->relation.mbrp_rev = -1;

	osm_tags_uninit(&obj->relation.curr->tags);
	for(n = 0; n < obj->relation.curr->members.used; n++)
		osm_relmember_free_fields(&obj->relation.curr->members.array[n]);
	vtr0_uninit(&obj->relation.curr->members);

	if (obj->relation.curr != &obj->relation.orig) {
		free(obj->relation.curr);
		obj->relation.curr = &obj->relation.orig;
	}
}

static void osm_relation_bbox(osm_any_obj_t *obj)
{
		obj->relation.hdr.bbox.x1 = obj->relation.hdr.bbox.y1 = +RND_COORD_MAX;
		obj->relation.hdr.bbox.x2 = obj->relation.hdr.bbox.y2 = -RND_COORD_MAX;

	if (obj->relation.curr->members.used > 0) {
		int n;
		osm_layer_t *ly = obj->relation.hdr.layer;

		for(n = 0; n < obj->relation.curr->members.used; n++) {
			osm_any_obj_t *mbo = osm_relmember_get_obj(ly, &obj->relation.curr->members.array[n]);
			if (mbo != NULL)
				osm_rtree_box_bump(&obj->relation.hdr.bbox, osm_obj_bbox(mbo));
		}
	}

	obj->relation.hdr.bbox_valid = 1;
}

/* Re-generate the mbrp cache if the layer has changed since last generated */
RND_INLINE void osm_relation_ensure_mbrp_(osm_relation_t *relation)
{
	long n;
	osm_layer_t *ly = relation->hdr.layer;

	if (!relation->hdr.dirty && (relation->mbrp_rev >= ly->rev))
		return;

	/* need to re-generate the cache */
	relation->mbrp.used = 0;
	for(n = 0; n < relation->curr->members.used; n++) {
		osm_any_obj_t *mbo = osm_relmember_get_obj(ly, &relation->curr->members.array[n]);
		if (mbo == NULL)
			rnd_message(RND_MSG_ERROR, "osm_relation_ensure_mbrp_(): invalid id %d:%lu\n", relation->curr->members.array[n].type, (unsigned long)relation->curr->members.array[n].ref);
		else
			vtp0_append(&relation->mbrp, mbo);
	}

	relation->mbrp_rev = ly->rev;

	if (relation->hdr.dirty) {
		osm_any_obj_t *o = (osm_any_obj_t *)relation;
		osm_objcom_rtree_unreg(o);
		o->relation.hdr.bbox_valid = 0;
		osm_relation_bbox(o);
		osm_objcom_rtree_reg(o);

		relation->hdr.dirty = 0;
	}
}

void osm_relation_ensure_mbrp(osm_relation_t *relation)
{
	osm_relation_ensure_mbrp_(relation);
}

static void osm_relation_draw(osm_any_obj_t *obj, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform, int overlay)
{
	TODO("probably can not draw relations");
}

static void osm_relation_draw_attached(osm_any_obj_t *obj, rnd_hid_gc_t gc)
{
	TODO("call low level rel draw");
}


static osm_any_obj_t *osm_relation_alloc(void) { return (osm_any_obj_t *)osm_relation_new(); }

static int osm_relation_isc_box(const osm_any_obj_t *o, const osm_rtree_box_t *box)
{
	TODO("probably can not be intersected");
	return 0;
}

static osm_tags_t *osm_relation_get_tags(osm_any_obj_t *o)
{
	return &o->relation.curr->tags;
}

void osm_relation_data_copy(osm_relation_data_t *dst, const osm_relation_data_t *src, int deep)
{
	if (deep) {
		long n;
		osm_relmember_t *s;

		for(n = 0,s = src->members.array; n < src->members.used; n++,s++) {
			osm_relmember_t *d = vtr0_alloc_append(&dst->members, 1);
			d->type = s->type;
			d->ref = s->ref;
			d->role = (s->role == NULL ? NULL : rnd_strdup(s->role));
		}
		osm_tags_dup(&dst->tags, &src->tags);
	}
}

static void osm_relation_members_uninit(osm_relation_data_t *rd)
{
	long n;
	osm_relmember_t *s;

	for(n = 0,s = rd->members.array; n < rd->members.used; n++,s++)
		osm_relmember_free_fields(s);
	vtr0_uninit(&rd->members);
}

static void osm_relation_fork(osm_relation_t *rel)
{
	assert(rel->curr == &rel->orig);

	rel->curr = malloc(sizeof(osm_relation_data_t));
	osm_relation_data_copy(rel->curr, &rel->orig, 1);
}

static void osm_relation_unfork(osm_relation_t *rel)
{
	assert(rel->curr != &rel->orig);
	osm_relation_members_uninit(rel->curr);
	osm_tags_uninit(&rel->curr->tags);
	free(rel->curr);
	rel->curr = &rel->orig;
}

static void osm_relation_pre_edit(osm_any_obj_t *o, int geo, int auto_fork)
{
	if (auto_fork && (o->relation.curr == &o->relation.orig))
		osm_relation_fork(&o->relation);
	if (geo) {
		osm_objcom_rtree_unreg(o);
		o->relation.hdr.bbox_valid = 0;
	}
}

static void osm_relation_post_edit(osm_any_obj_t *o, int tags_changed)
{
	o->relation.mbrp_rev = -1;

	if (!o->relation.hdr.bbox_valid) {
		osm_relation_bbox(o);
		osm_rtree_insert(&o->hdr.layer->objs, o, &o->hdr.bbox);
	}
	if (tags_changed)
		osm_style_update_obj(o);
	o->hdr.layer->rev++;
}


static void osm_relation_referee_objs(osm_any_obj_t *o, vtp0_t *dst, int update_att)
{
	/* Not recursive. */
}

static long osm_relation_referee_idx(osm_any_obj_t *parent, osm_any_obj_t *child)
{
	long n;
	osm_relation_t *rel = (osm_relation_t *)parent;
	osm_relmember_t *s;

	for(n = 0,s = rel->curr->members.array; n < rel->curr->members.used; n++,s++)
		if ((s->type == child->hdr.type) && (s->ref == child->hdr.id))
			return n;

	return -1;
}

static void osm_relation_diffresult(osm_any_obj_t *o, osm_id_t new_id, osm_id_t new_version)
{
	osm_tags_uninit(&o->relation.orig.tags);
	osm_relation_members_uninit(&o->relation.orig);
	osm_relation_data_copy(&o->relation.orig, o->relation.curr, 0);
	o->relation.orig.tags = o->relation.curr->tags;
	o->relation.orig.members = o->relation.curr->members;
	free(o->relation.curr);
	o->relation.curr = &o->relation.orig;
	osm_objcom_diffresult(o, new_id, new_version);

}

static void osm_relation_diffresult_referees(osm_any_obj_t *o, htdr_t *dr, int permit_del, int undoable)
{
	TODO("implement me");
}

static void osm_relation_pre_remove_referee(osm_any_obj_t *o, osm_objtype_t rem_ty, osm_id_t rem_id, int undoable)
{
	TODO("implement me");
}

static void osm_relation_post_remove(osm_any_obj_t *o, int undoable)
{
	/* nothing to do: orphaned objects become normal objects */
}


osm_any_obj_t *osm_relation_dup_into(osm_any_obj_t *o, osm_layer_t *dst, osm_id_t new_id)
{
	osm_any_obj_t *obj = osm_relation_alloc();

	obj->relation.mbrp_rev = -1;
	osm_relation_data_copy(obj->relation.curr, o->relation.curr, 1);
	obj->relation.mbrp_rev = -1;
	if (new_id != 0)
		obj->hdr.id = new_id;
	else
		obj->hdr.id = o->hdr.id;
	osm_obj_add_to_layer(dst, obj);
	return obj;
}

static const char *osm_relation_visual_type(osm_any_obj_t *o)
{
	return htsp_get(&o->relation.curr->tags, "type");
}

static const osm_objcalls_t osm_relation_calls = {
	osm_relation_alloc,
	osm_relation_free_fields,
	osm_relation_draw,
	osm_relation_draw_attached,
	osm_relation_bbox,
	osm_relation_isc_box,
	osm_relation_get_tags,
	osm_relation_pre_edit,
	osm_relation_post_edit,
	osm_relation_referee_objs,
	osm_relation_referee_idx,
	osm_relation_diffresult,
	osm_relation_diffresult_referees,
	osm_relation_pre_remove_referee,
	osm_relation_post_remove,
	osm_relation_dup_into,
	osm_relation_visual_type
};

void osm_relation_init(osm_relation_t *relation)
{
	memset(relation, 0, sizeof(osm_relation_t));
	osm_objcom_init(&relation->hdr, OSM_OBJ_RELATION, &osm_relation_calls);
	relation->curr = &relation->orig;
	osm_tags_init(&relation->orig.tags);
	relation->mbrp_rev = -1;
}

osm_relation_t *osm_relation_new(void)
{
	osm_relation_t *res = malloc(sizeof(osm_relation_t));
	osm_relation_init(res);
	return res;
}


int osm_relation_refers_obj(osm_relation_t *rel, osm_any_obj_t *obj)
{
	long n;
	osm_layer_t *ly = rel->hdr.layer;

	for(n = 0; n < rel->curr->members.used; n++) {
		osm_relmember_t *rm = &(rel->curr->members.array[n]);
		osm_any_obj_t *referee = htIp_get(&(ly->id2obj[rm->type]), rm->ref);
		if ((referee != NULL) && (referee == obj))
			return 1;
	}

	return 0;
}

long osm_relation_list_referers_by_obj(vtp0_t *lst, osm_any_obj_t *obj)
{
	osm_rtree_it_t it;
	void *o;
	long count = 0;

	for(o = osm_rtree_first(&it, &obj->hdr.layer->objs, &obj->hdr.bbox); o != NULL; o = osm_rtree_next(&it)) {
		osm_relation_t *rel = o;
		if ((rel->hdr.type == OSM_OBJ_RELATION) && !rel->hdr.referer && osm_relation_refers_obj(rel, obj)) {
			vtp0_append(lst, rel);
			rel->hdr.referer = 1;
			count++;
		}
	}
	return count;
}


long osm_relation_dirt_referers_by_obj(osm_any_obj_t *obj)
{
	osm_rtree_it_t it;
	void *o;
	long count = 0;

	for(o = osm_rtree_first(&it, &obj->hdr.layer->objs, &obj->hdr.bbox); o != NULL; o = osm_rtree_next(&it)) {
		osm_relation_t *rel = o;
		if ((rel->hdr.type == OSM_OBJ_RELATION) && !rel->hdr.referer && osm_relation_refers_obj(rel, obj)) {
			rel->hdr.dirty = 1;
			count++;
		}
	}
	return count;
}

#include "obj_relation_undo.c"
