/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/core/config.h>
#include <librnd/core/actions.h>
#include <librnd/hid/hid.h>

#include "conf_core.h"
#include "data.h"
#include "obj_any.h"
#include "search.h"
#include "draw.h"
#include "event.h"
#include "funchash_core.h"

#include "select.h"

void osm_obj_unselect(osm_any_obj_t *o)
{
	o->hdr.selected = 0;
}

void osm_obj_select(osm_any_obj_t *o)
{
	o->hdr.selected = 1;
}

static void sel_chg_notify(osm_layer_t *ly)
{
	rnd_event(&ly->parent->hidlib, OSM_EVENT_SELECTION_CHANGED, NULL);
}


void osm_unselect_all(osm_layer_t *ly)
{
	htIp_entry_t *e;
	int t;

	for(t = 0; t < OSM_OBJ_max; t++)
		for(e = htIp_first(&ly->id2obj[t]); e != NULL; e = htIp_next(&ly->id2obj[t], e))
			osm_obj_unselect(e->value);

	sel_chg_notify(ly);
}

void osm_select_click(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y)
{
	osm_any_obj_t *obj;
	
	if (rnd_gui->shift_is_pressed(rnd_gui))
		obj = osm_search_obj_xy(ly, x, y); /* for the shift+click unselect */
	else
		obj = osm_search_obj_xy_unsel(ly, x, y);

	if (obj == NULL) { /* clicked away or clicked on selected */
		if (!rnd_gui->shift_is_pressed(rnd_gui))
			osm_unselect_all(ly);
		return;
	}

	if (obj->hdr.selected) {
		if (rnd_gui->shift_is_pressed(rnd_gui))
			osm_obj_unselect(obj);
		else
			osm_obj_select(obj);
	}
	else {
		if (!rnd_gui->shift_is_pressed(rnd_gui))
			osm_unselect_all(ly);
		osm_obj_select(obj);
	}
	sel_chg_notify(ly);
}


void osm_select_box(osm_layer_t *ly, rnd_coord_t x1, rnd_coord_t y1, rnd_coord_t x2, rnd_coord_t y2)
{
	int neg = 0;
	vtp0_t res = {0};
	rnd_coord_t slop = osm_slop;
	osm_rtree_box_t rbox;
	long n;

	osm_draw_inhibit_inc();

	if (!rnd_gui->shift_is_pressed(rnd_gui))
		osm_unselect_all(ly);

	/* figure if selection is negative */
	if (!conf_core.editor.selection.disable_negative) {
		int negx = x2 < x1, negy = y1 < y2; /* y is inverted on screen */

		if (!conf_core.editor.selection.symmetric_negative)
			neg = negy;

		neg |= negx;
	}

	/* make the box canonical */
	if (x1 > x2) rnd_swap(rnd_coord_t, x1, x2);
	if (y1 > y2) rnd_swap(rnd_coord_t, y1, y2);

	rbox.x1 = x1 - slop; rbox.y1 = y1 - slop;
	rbox.x2 = x2 + slop; rbox.y2 = y2 + slop;
	osm_search_append_objs_box(&res, ly, &rbox);

	for(n = 0; n < res.used; n++) {
		osm_any_obj_t *o = res.array[n];

		if (neg) { /* select anything that intersects */
			if (o->hdr.calls->isc_box(o, &rbox))
				osm_obj_select(o);
		}
		else { /* select anything that is fully within */
			if ((o->hdr.bbox.x1 >= rbox.x1) && (o->hdr.bbox.y1 >= rbox.y1) && (o->hdr.bbox.x2 <= rbox.x2) && (o->hdr.bbox.y2 <= rbox.y2))
				osm_obj_select(o);
		}
	}

	sel_chg_notify(ly);

	vtp0_uninit(&res);

	osm_draw_inhibit_dec();
}

void osm_select_all_mask_tag(osm_layer_t *ly, osm_objtype_mask_t mask, osm_select_all_type_t restrictions, vtp0_t *keyval)
{
	htIp_entry_t *e;
	int t, n;

	for(t = 0; t < OSM_OBJ_max; t++) {
		osm_objtype_mask_t tm = 1 << t;

		if (!(mask & tm))
			continue;

		for(e = htIp_first(&ly->id2obj[t]); e != NULL; e = htIp_next(&ly->id2obj[t], e)) {
			osm_tags_t *tags = NULL;
			osm_any_obj_t *obj = (osm_any_obj_t *)e->value;

			if ((restrictions & OSM_SELAT_FRESH) ) {
				if (obj->hdr.id > 0)
					continue; /* not new */
				tags = obj->hdr.calls->get_tags(obj);
				if (tags->used > 0)
					continue; /* tagged */
			}

			if ((keyval != NULL) && (keyval->used != 0)) {
				int bad = 0;

				if (tags == NULL)
					tags = obj->hdr.calls->get_tags(obj);

				if (tags == NULL)
					continue; /* can not match because obj has no tags */

				for(n = 0; n < keyval->used; n+=2) {
					static const char *empty = "";
					const char *want_key = keyval->array[n];
					const char *want_val = keyval->array[n+1], *val;
					int want_empty = (want_val == NULL) || (*want_val == '\0');

					val = htsp_get(tags, want_key);
					if (val == NULL)
						val = empty;

					/* key doesn't exist or is empty; refuse if want_val is non-empty */
					if ((*val == '\0') && !want_empty)
						{ bad = 1; break; }

					/* obj's value is non-empty but caller wants empty: refuse */
					if (want_empty)
						{ bad = 1; break; }

					/* by now neither val nor want_val is NULL; refuse on val mismatch */
					if (strcmp(val, want_val) != 0)
						{ bad = 1; break; }
				}
				if (bad)
					continue;
			}

			osm_obj_select(obj);
		}
	}

	sel_chg_notify(ly);

}

void osm_select_all_mask(osm_layer_t *ly, osm_objtype_mask_t mask, osm_select_all_type_t restrictions)
{
	osm_select_all_mask_tag(ly, mask, restrictions, NULL);
}


static const char osm_acts_Unselect[] = "Unselect(All)";
static const char osm_acth_Unselect[] = "Unselect all objects\n";
fgw_error_t osm_act_Unselect(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);

	osm_unselect_all(ly);

	RND_ACT_IRES(0);
	return 0;
}

static const char osm_acts_Select[] = "Select([all|fresh], [node|way], [tag,val, tag,val, ...])\n";
static const char osm_acth_Select[] = "Adds objects to selection. Fresh selects only new, untagged objects. Second argument can be used to filter for object type. If tags are specified select objects that matches all tags (AND relation assumed)";
static fgw_error_t osm_act_Select(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	int op, n;
	osm_objtype_mask_t mask = OSM_MASK_NODE | OSM_MASK_WAY;
	const char *smask = NULL;
	vtp0_t tag_match = {0};

	RND_ACT_CONVARG(1, FGW_KEYWORD, Select, op = fgw_keyword(&argv[1]));
	RND_ACT_MAY_CONVARG(2, FGW_STR, Select, smask = argv[2].val.str);

	if (smask != NULL) {
		if (strcmp(smask, "way") == 0) mask = OSM_MASK_WAY;
		else if (strcmp(smask, "node") == 0) mask = OSM_MASK_NODE;
		else if (strcmp(smask, "any") == 0) mask = OSM_MASK_NODE|OSM_MASK_WAY;
		else {
			rnd_message(RND_MSG_ERROR, "Invalid object mask '%s'\n", smask);
			return FGW_ERR_ARG_CONV;
		}
	}

	for(n = 3; n < argc; n+=2) {
		char *want_key, *want_val;
		RND_ACT_MAY_CONVARG(n+0, FGW_STR, Select, want_key = argv[n+0].val.str);
		RND_ACT_MAY_CONVARG(n+1, FGW_STR, Select, want_val = argv[n+1].val.str);
		vtp0_append(&tag_match, want_key);
		vtp0_append(&tag_match, want_val);
	}

	switch(op) {
		case F_All:
			osm_select_all_mask_tag(ly, mask, 0, &tag_match);
			break;
		case F_Fresh:
			osm_select_all_mask_tag(ly, mask, OSM_SELAT_FRESH, &tag_match);
			break;
	}

	vtp0_uninit(&tag_match);

	return 0;
}


static rnd_action_t select_action_list[] = {
	{"Unselect", osm_act_Unselect, osm_acth_Unselect, osm_acts_Unselect},
	{"Select", osm_act_Select, osm_acth_Select, osm_acts_Select}
};


void osm_select_init2(void)
{
	RND_REGISTER_ACTIONS(select_action_list, NULL);
}

