/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/core/compat_misc.h>
#include <genht/hash.h>
#include "data.h"
#include "event.h"
#include "obj_any.h"
#include "obj_common.h"

#include "style.h"

vtp0_t osm_styles;
htsp_t osm_name2style;

osm_style_t *osm_def_node_style, *osm_def_way_style;

long osm_style_new(const char *name)
{
	osm_style_t *st;

	st = htsp_get(&osm_name2style, name);
	if (st != NULL)
		return st->id;

	st = calloc(sizeof(osm_style_t), 1);
	st->name = rnd_strdup(name);

	/* by default use low style data only */
	st->lod_invis_above = 1000000;
	st->lod_high_below = 0;

	st->id = osm_styles.used;
	vtp0_append(&osm_styles, st);
	htsp_set(&osm_name2style, st->name, st);

	return st->id;
}

static void osm_style_free(osm_style_t *st)
{
	free(st->name);
	free(st);
}

void osm_style_update_obj(osm_any_obj_t *o)
{
	osm_style_t *st = NULL;

	rnd_event(&osm.hidlib, OSM_EVENT_STYLE_DETERMINE, "pp", &st, o);

	/* fallback */
	if (st == NULL) {
		switch(o->hdr.type) {
			case OSM_OBJ_NODE: st = osm_def_node_style; break;
			case OSM_OBJ_WAY:  st = osm_def_way_style; break;
			case OSM_OBJ_RELATION:
			case OSM_OBJ_CHANGESET:
			case OSM_OBJ_invalid:
			case OSM_OBJ_vmax:
				/* do not render */
				break;
		}
	}

	/* re-register in the zorder rtrees if zorder has changed */
	if ((o->hdr.style != st) && ((st == NULL) || (o->hdr.style == NULL) || (o->hdr.style->zord != st->zord))) {
		osm_objcom_rtree_unreg_style(o);
		o->hdr.style = st;
		osm_objcom_rtree_reg_style(o);
	}
}


void osm_style_system_init(void)
{
	long id;

	htsp_init(&osm_name2style, strhash, strkeyeq);

	id = osm_style_new("default node fallback");
	osm_def_node_style = osm_styles.array[id];
	osm_def_node_style->low.stroke.color = *rnd_color_golden;
	osm_def_node_style->low.stroke.width = -2;
	osm_def_node_style->high.stroke.color = *rnd_color_golden;
	osm_def_node_style->high.stroke.width = -3;
	osm_def_node_style->lod_invis_above = 3;
	osm_def_node_style->lod_high_below = 1;
	osm_def_node_style->fallback = 1;

	id = osm_style_new("default way fallback");
	osm_def_way_style = osm_styles.array[id];
	osm_def_way_style->low.stroke.color = *rnd_color_magenta;
	osm_def_way_style->low.stroke.width = -1;
	osm_def_way_style->fallback = 1;
}

void osm_style_system_uninit(void)
{
	long n;
	htsp_uninit(&osm_name2style);
	for(n = 0; n < osm_styles.used; n++)
		osm_style_free(osm_styles.array[n]);
	vtp0_uninit(&osm_styles);
}

