/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/config.h>

#include <stdlib.h>
#include <string.h>

#include <librnd/core/actions.h>
#include <librnd/hid/hid_dad.h>
#include <librnd/hid/hid_dad_tree.h>

#include "data.h"
#include "obj_any.h"
#include "select.h"
#include "search.h"
#include "funchash_core.h"

#include "dlg_props.h"

typedef struct{
	RND_DAD_DECL_NOINIT(dlg)
	int wreferees, wusedby, wlat, wlon;

	/* object address */
	osm_layer_t *ly;
	osm_objtype_t obj_type;
	osm_id_t obj_id;

	/* temporary values for listing */
	long cnt;
} props_ctx_t;

static void props_close_cb(void *caller_data, rnd_hid_attr_ev_t ev)
{
	props_ctx_t *ctx = caller_data;

	RND_DAD_FREE(ctx->dlg);
	free(ctx);
}

static void props_osm2dlg_node(props_ctx_t *ctx, osm_any_obj_t *obj)
{
	TODO("fill in node coords");
}

static void props_osm2dlg_referees(props_ctx_t *ctx, osm_any_obj_t *obj)
{
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	rnd_hid_row_t *r, *row;
	char *cell[3], *cursor_path = NULL;
	rnd_hid_attr_val_t hv;
	long n;

	/* reload the list */
	attr = &ctx->dlg[ctx->wreferees];
	tree = attr->wdata;

	/* remember cursor */
	r = rnd_dad_tree_get_selected(attr);
	if (r != NULL)
		cursor_path = rnd_strdup(r->cell[0]);

	/* remove existing items */
	rnd_dad_tree_clear(tree);

	cell[2] = NULL;

	switch(obj->hdr.type) {
		case OSM_OBJ_WAY:
			for(n = 0; n < obj->way.curr->nd.used; n++) {
				cell[0] = rnd_strdup_printf("%ld", n);
				cell[1] = rnd_strdup_printf("node #" OSM_PID, obj->way.curr->nd.array[n]);
				row = rnd_dad_tree_append(attr, NULL, cell);
				row->user_data2.lng = n;
			}
			break;
		case OSM_OBJ_RELATION:
			rnd_message(RND_MSG_ERROR, "TODO: not yet implemented for relations\n");
			break;
		case OSM_OBJ_NODE:
		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_vmax:
		case OSM_OBJ_invalid:
			break;
		}



	/* restore cursor */
	if (cursor_path != NULL) {
		hv.str = cursor_path;
		rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wreferees, &hv);
	}

	free(cursor_path);
}

static void way_referer_cb(void *uctx, osm_way_t *way, long member_idx)
{
	char *cell[6];
	props_ctx_t *ctx = uctx;
	rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wusedby];
	const char *type = osm_obj_visual_type((osm_any_obj_t *)way);

	if (type == NULL)
		type = "";

	cell[0] = rnd_strdup_printf("%ld", ctx->cnt);
	cell[1] = rnd_strdup_printf("way #" OSM_PID " %s", way->hdr.id, type);
	cell[2] = rnd_strdup_printf("%ld", member_idx);
	cell[3] = NULL;
	rnd_dad_tree_append(attr, NULL, cell);

	ctx->cnt++;
}


static void rel_referer_cb(void *uctx, osm_relation_t *rel, long member_idx)
{
	char *cell[6];
	props_ctx_t *ctx = uctx;
	rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wusedby];
	osm_relmember_t *rm = rel->curr->members.array+member_idx;
	char *type;

	type = htsp_get(&rel->curr->tags, "type");
	if (type == NULL)
		type = "";

	cell[0] = rnd_strdup_printf("%ld", ctx->cnt);
	cell[1] = rnd_strdup_printf("relation #" OSM_PID " %s", rel->hdr.id, type);
	cell[2] = rnd_strdup_printf("%ld", member_idx);
	cell[3] = rnd_strdup(rm->role == NULL ? "" : rm->role);
	cell[4] = NULL;
	rnd_dad_tree_append(attr, NULL, cell);

	ctx->cnt++;
}


static void props_osm2dlg_referers(props_ctx_t *ctx, osm_any_obj_t *obj)
{
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	rnd_hid_row_t *r;
	char *cursor_path = NULL;
	rnd_hid_attr_val_t hv;

	/* reload the list */
	attr = &ctx->dlg[ctx->wusedby];
	tree = attr->wdata;

	/* remember cursor */
	r = rnd_dad_tree_get_selected(attr);
	if (r != NULL)
		cursor_path = rnd_strdup(r->cell[0]);

	/* remove existing items */
	rnd_dad_tree_clear(tree);


	if (obj != NULL) {
		ctx->cnt = 0;
		if (obj->hdr.type == OSM_OBJ_NODE)
			osm_search_referer_ways(&obj->node, way_referer_cb, ctx);
		osm_search_referer_relations(obj, rel_referer_cb, ctx);
	}


	/* restore cursor */
	if (cursor_path != NULL) {
		hv.str = cursor_path;
		rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wreferees, &hv);
	}

	free(cursor_path);
}

static void props_osm2dlg_usedby(props_ctx_t *ctx, osm_any_obj_t *obj)
{
	TODO("implement me");
}

static void props_osm2dlg(props_ctx_t *ctx)
{
	osm_any_obj_t *obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if (ctx->obj_type == OSM_OBJ_NODE)
		props_osm2dlg_node(ctx, obj);
	else
		props_osm2dlg_referees(ctx, obj);
	props_osm2dlg_referers(ctx, obj);
	props_osm2dlg_usedby(ctx, obj);
}


/*
static void referee_select(rnd_hid_attribute_t *attrib, void *hid_ctx, rnd_hid_row_t *row)
{
	rnd_hid_tree_t *tree = attrib->wdata;
	props_ctx_t *ctx = tree->user_ctx;
	const char *key = NULL, *val = NULL;
	rnd_hid_attr_val_t hv;

	if (row != NULL) {
		key = row->cell[0];
		val = row->cell[1];
	}

	TODO("set suggestions");

	hv.str = val;
	rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wval, &hv);
}
*/


static void cb_main_zoomto(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	props_ctx_t *ctx = caller_data;
	osm_design_t *osm = ctx->ly->parent;
	fgw_arg_t args[5] = {0}, res;
	char tmp[128];

	sprintf(tmp, OSM_PID, ctx->obj_id);

	args[1].type = FGW_STR; args[1].val.cstr = "obj";
	args[2].type = FGW_STR; args[2].val.cstr = osm_obj_name[ctx->obj_type];
	args[3].type = FGW_STR; args[3].val.cstr = tmp;

	rnd_actionv_bin(&osm->hidlib, "zoom", &res, 4, args);
}

/*	osm_design_t *osm = ctx->ly->parent; \*/
#define GET_ROW \
	props_ctx_t *ctx = caller_data; \
	rnd_hid_attribute_t *lst_attr = &ctx->dlg[ctx->wreferees]; \
	rnd_hid_row_t *row; \
 \
	row = rnd_dad_tree_get_selected(lst_attr); \
	if (row == NULL) { \
		rnd_message(RND_MSG_ERROR, "Need to select a referee first\n"); \
		return; \
	}

static void cb_referee_zoomto(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	fgw_arg_t args[5] = {0}, res;
	GET_ROW;

	{
		osm_design_t *osm = ctx->ly->parent;

		args[1].type = FGW_STR; args[1].val.cstr = "obj";
		args[2].type = FGW_STR; args[2].val.cstr = row->cell[1];

		rnd_actionv_bin(&osm->hidlib, "zoom", &res, 3, args);
	}
}


static void cb_main_select(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	props_ctx_t *ctx = caller_data;
/*	osm_design_t *osm = ctx->ly->parent;*/
	osm_any_obj_t *obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if (obj != NULL) {
		osm_obj_select(obj);
		rnd_gui->invalidate_all(rnd_gui);
	}
}


static void cb_main_unselect(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	props_ctx_t *ctx = caller_data;
/*	osm_design_t *osm = ctx->ly->parent;*/
	osm_any_obj_t *obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if (obj != NULL) {
		osm_obj_unselect(obj);
		rnd_gui->invalidate_all(rnd_gui);
	}
}

static void cb_referee_select(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	osm_any_obj_t *obj;
	osm_objtype_t type;
	osm_id_t id;
	GET_ROW;

	type = osm_str_to_obj_type_id(row->cell[1], NULL, &id);
	if (type == OSM_OBJ_invalid) {
		rnd_message(RND_MSG_ERROR, "Internal error: invalid type:#id pair\n");
		return;
	}
	
	obj = osm_obj_by_id(ctx->ly, type, id);
	if (obj != NULL) {
		osm_obj_select(obj);
		rnd_gui->invalidate_all(rnd_gui);
	}
}

static void cb_referee_unselect(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	osm_any_obj_t *obj;
	osm_objtype_t type;
	osm_id_t id;
	GET_ROW;

	type = osm_str_to_obj_type_id(row->cell[1], NULL, &id);
	if (type == OSM_OBJ_invalid) {
		rnd_message(RND_MSG_ERROR, "Internal error: invalid type:#id pair\n");
		return;
	}
	
	obj = osm_obj_by_id(ctx->ly, type, id);
	if (obj != NULL) {
		osm_obj_unselect(obj);
		rnd_gui->invalidate_all(rnd_gui);
	}
}

static void cb_referee_remove_lst(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	osm_any_obj_t *obj;
	GET_ROW;


	obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);
	if (obj == NULL)
		return;

	switch(obj->hdr.type) {
		case OSM_OBJ_WAY:
			osm_way_modify_remove(&obj->way, row->user_data2.lng, 1);
			break;
		case OSM_OBJ_RELATION:
			rnd_message(RND_MSG_ERROR, "TODO: not yet implemented for relations\n");
			break;
		case OSM_OBJ_NODE:
		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_vmax:
		case OSM_OBJ_invalid:
			rnd_message(RND_MSG_ERROR, "Internal error: can't get here in cb_referee_remove_lst()\n");
			return;
	}

	rnd_gui->invalidate_all(rnd_gui);
	props_osm2dlg(ctx);
}

static int osm_rnd_dlg_props(osm_layer_t *ly, osm_objtype_t type, osm_id_t id)
{
	rnd_hid_dad_buttons_t clbtn[] = {{"Close", 0}, {NULL, 0}};
	const char *rhdr[] = {"#", "referee"};
	const char *uhdr[] = {"#", "referer", "place", "role", "comment"};
	const char *tabs[] = {NULL, "Used by", NULL};
	char title[256];
	props_ctx_t *ctx;
	int modal = 0, res = 0, is_node = (type == OSM_OBJ_NODE);
	osm_any_obj_t *obj = osm_obj_by_id(ly, type, id);

	if (obj == NULL)
		return -1;

	tabs[0] = is_node ? "Coords" : "Referees";

	ctx = calloc(sizeof(props_ctx_t), 1);
	ctx->ly = ly;
	ctx->obj_type = type;
	ctx->obj_id = id;

	RND_DAD_BEGIN_VBOX(ctx->dlg);
		RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

		RND_DAD_BEGIN_TABBED(ctx->dlg, tabs);
			RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* First tab: referees/coords */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				if (!is_node) {
					RND_DAD_TREE(ctx->dlg, 2, 0, rhdr);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
						ctx->wreferees = RND_DAD_CURRENT(ctx->dlg);
					RND_DAD_BEGIN_HBOX(ctx->dlg);
						RND_DAD_LABEL(ctx->dlg, "Referee:");
						RND_DAD_BUTTON(ctx->dlg, "zoom to");
							RND_DAD_CHANGE_CB(ctx->dlg, cb_referee_zoomto);
						RND_DAD_BUTTON(ctx->dlg, "select");
							RND_DAD_CHANGE_CB(ctx->dlg, cb_referee_select);
						RND_DAD_BUTTON(ctx->dlg, "unselect");
							RND_DAD_CHANGE_CB(ctx->dlg, cb_referee_unselect);
						RND_DAD_BUTTON(ctx->dlg, "remove from list");
							RND_DAD_CHANGE_CB(ctx->dlg, cb_referee_remove_lst);
					RND_DAD_END(ctx->dlg);
				}
				else {
					RND_DAD_BEGIN_VBOX(ctx->dlg);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);
						RND_DAD_BEGIN_TABLE(ctx->dlg, 2);
							RND_DAD_LABEL(ctx->dlg, "Lat:");
							RND_DAD_REAL(ctx->dlg);
								ctx->wlat = RND_DAD_CURRENT(ctx->dlg);
							RND_DAD_LABEL(ctx->dlg, "Lon:");
							RND_DAD_REAL(ctx->dlg);
								ctx->wlon = RND_DAD_CURRENT(ctx->dlg);
						RND_DAD_END(ctx->dlg);
					RND_DAD_END(ctx->dlg);
				}
				RND_DAD_BEGIN_HBOX(ctx->dlg);
					RND_DAD_LABEL(ctx->dlg, "Main object:");
					RND_DAD_BUTTON(ctx->dlg, "zoom to");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_main_zoomto);
					RND_DAD_BUTTON(ctx->dlg, "select");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_main_select);
					RND_DAD_BUTTON(ctx->dlg, "unselect");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_main_unselect);
				RND_DAD_END(ctx->dlg);


			RND_DAD_END(ctx->dlg);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* Second tab: referers */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);
				RND_DAD_TREE(ctx->dlg, 4, 0, uhdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wusedby = RND_DAD_CURRENT(ctx->dlg);
			RND_DAD_END(ctx->dlg);

		RND_DAD_END(ctx->dlg);


		RND_DAD_BUTTON_CLOSES(ctx->dlg, clbtn);
	RND_DAD_END(ctx->dlg);


	rnd_snprintf(title, sizeof(title), "properties of %s #" OSM_PID, osm_obj_name[obj->hdr.type], obj->hdr.id);
	RND_DAD_NEW("props", ctx->dlg, title, ctx, modal, props_close_cb);
	props_osm2dlg(ctx);

	if (modal)
		res = RND_DAD_RUN(ctx->dlg);

	return res;
}

const char osm_acts_PropsDialog[] = "PropsDialog(object)\n";
const char osm_acth_PropsDialog[] = "Present the PropsDialog box of an object";
fgw_error_t osm_act_PropsDialog(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	osm_any_obj_t *obj;
	int op;

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer\n");
		return -1;
	}

	RND_ACT_CONVARG(1, FGW_KEYWORD, PropsDialog, op = fgw_keyword(&argv[1]));
	switch(op) {
		case F_Object:
			obj = osm_search_obj_xy(ly, osm->crosshair_x, osm->crosshair_y);
			if (obj == NULL) {
				rnd_message(RND_MSG_ERROR, "No object under cursor\n");
				return -1;
			}
			break;
		case F_BinObject:
			RND_ACT_CONVARG(2, FGW_PTR, PropsDialog, obj = argv[2].val.ptr_void);
			break;
		default:
			rnd_message(RND_MSG_ERROR, "PropsDialog(): Invalid first argument\n");
			return -1;
	}

	osm_rnd_dlg_props(ly, obj->hdr.type, obj->hdr.id);
	RND_ACT_IRES(0);
	return 0;
}
