/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - TMS map tiles
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <math.h>
#include <genvector/vts0.h>
#include <librnd/config.h>
#include <librnd/core/hidlib.h>
#include <librnd/core/conf_multi.h>
#include <librnd/core/actions.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
#include <librnd/hid/pixmap.h>

#include "data.h"
#include "event.h"
#include "projection.h"

#include "img_tms_conf.h"

conf_tms_t conf_tms;

static const char img_tms_cookie[] = "img_tms";

#include "conf_internal.c"
#include "src_plugins/lib_tilecache/lib_tilecache.h"

/*** url support ***/
typedef struct tms_url_seg_s tms_url_seg_t;

struct tms_url_seg_s {
	enum {
		TUS_STR,      /* static string, use ->str */
		TUS_X,
		TUS_Y,
		TUS_Z,       /* zoom level */
		TUS_SWITCH   /* pick one of ->sw */
	} type;
	union {
		char *str;
		vts0_t sw;
	} data;
	int cnt; /* for sw */
	tms_url_seg_t *next;
};

#define NEWSEG(ty) \
do { \
	seg = calloc(sizeof(tms_url_seg_t), 1); \
	seg->type = ty; \
	if (head == NULL)   { head = tail = seg; } \
	else                { tail->next = seg; tail = seg; } \
} while(0)

RND_INLINE void split_switch(tms_url_seg_t *seg, const char *str)
{
	const char *curr, *next;

	for(curr = str; curr != NULL; curr = next) {
		char *tmp;

		next = strpbrk(curr, ",}");
		if (next != NULL) {
			tmp = rnd_strndup(curr, next-curr);
			if (*next == '}')
				next = NULL;
			else
				next++;
		}
		else
			tmp = rnd_strdup(curr);

		vts0_append(&seg->data.sw, tmp);
	}
}

typedef struct tms_s {
	char *name, *url;
	tms_url_seg_t *url_segs; /* parsed url */
	osm_tilecache_t cache;
	int maxlev;
	struct {
		rnd_coord_t x, y;
	} offs;
} tms_t;

static tms_url_seg_t *tms_url_parse(tms_t *tms, const char *url)
{
	const char *curr, *next;
	tms_url_seg_t *head = NULL, *tail = NULL, *seg;

	if (strncmp(url, "tms", 3) == 0) {
		url+=3;
		if (*url == '[') {
			char *end;
			long z;

			url++;
			z = strtol(url, &end, 10);
			if ((*end != ']') || (z < 2) || (z > 30)) {
				rnd_message(RND_MSG_ERROR, "Invalid tms [] max zoom level: must be an integer between 2 and 30\n");
				return NULL;
			}
			tms->maxlev = z;
			url = end+1; /* skip over "]" */
		}
		if (*url == ':')
			url++;
	}

	for(curr = url; curr != NULL; curr = next) {
		/* str context */
		if (*curr == '\0')
			break;
		next = strchr(curr, '{');
		if (next == NULL) {
			make_str:;
			NEWSEG(TUS_STR);
			seg->data.str = rnd_strdup(curr);
		}
		else {
			NEWSEG(TUS_STR);
			seg->data.str = rnd_strndup(curr, next-curr);
			curr = next+1;
			next = strchr(curr, '}');
			if (next == NULL) {
				rnd_message(RND_MSG_ERROR, "TMS: unterminated {} field at %s\n", curr);
				goto make_str;
			}
			switch(*curr) {
				case 'x': NEWSEG(TUS_X); break;
				case 'y': NEWSEG(TUS_Y); break;
				case 'z': NEWSEG(TUS_Z); break;
				case 's':
					if (strncmp(curr, "switch:", 7) == 0) {
						curr += 7;
						NEWSEG(TUS_SWITCH);
						split_switch(seg, curr);
						break;
					}
				default:
					rnd_message(RND_MSG_ERROR, "TMS: invalid {} field at %s\n", curr);
			}
			next++;
		}
	}

	return head;
}


/*** imaging ***/

typedef struct tms_tileaddr_s {
	long x, y;
	int z;
} tms_tileaddr_t;

#include "cache.c"

#define PI     3.14159265358979323846  /* pi */
#define invPI  0.31830988618379067154  /* 1/pi */

double y2lat(int y, int z)
{
	double n = PI*(1.0 - (2.0*y+1.0)/(1<<z));
	return 180.0/PI * atan(sinh(n));
}


/* Convert lat:lon:level to tile coord x;y */
void tms_latlon2xy(double lon, double lat, int level, long *tx, long *ty)
{
	double zm = (1L<<level);
	double latrad = lat*PI/180.0;

	*tx = zm * (lon+180.0) / 360.0;
	*ty = zm * (1.0 - invPI * log(tan(latrad) + 1.0/cos(latrad))) / 2;
}

/* Convert tile-pixel cords to lat/lon */
void tms_pxy2latlon(double bpx, double bpy, int level, double *lon, double *lat)
{
	double zm = (1L<<level);
	double nlat = PI*(1.0 - (2.0*bpy+1.0)/zm);

	*lon = (360.0 * bpx + 180.0)/zm - 180.0;
	*lat = 180.0/PI * atan(sinh(nlat));
}

void tms_tile2key(char *dst, long tilex, long tiley, int level)
{
	sprintf(dst, "z%d_x%ld_y%ld", level, tilex, tiley);
}

/*** Render/draw ***/
static double tile_meter_per_pix[24] = {
	0, 78271.5170, 39135.7585, 19567.8792, 9783.9396, 4891.9698, 2445.9849,
	1222.9925, 611.4962, 305.7481, 152.8741, 76.4370, 38.2185, 19.1093, 9.5546,
	4.7773, 2.3887, 1.1943, 0.5972, 0.2986, 0.1493, 0.0746, 0.0373, 0.0187};

static void img_tms_layer_draw(osm_layer_t *ly, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform_caller)
{
	tms_t *tms = ly->plugin_data;
	double lon, lat, t0clon, t0clat;
	int n;
	int level = osm_zoom_rnd2level(region->coord_per_pix);
	long bpx, bpy;     /* TMS pixel coords (at a specific zoom level) */
	long tile0x, tile0y; /* TMS tile coords (at a specific zoom level) */
	rnd_coord_t t0cx, t0cy; /* left-most TMS tile's center point coords */
	long tilex, tiley;
	rnd_coord_t tcx, tcy;
	time_t now = time(NULL);
	rnd_design_t *dsg = (rnd_design_t *)ly->parent;
	char key[256];

	if (level < 1)
		return; /* not available */

	if (level > tms->maxlev)
		level = tms->maxlev;

	rnd2lldeg(&lon, &lat, region->view.X1, region->view.Y2);

	tms_latlon2xy(lon, lat, level, &tile0x, &tile0y);

	/* center of the first tile */
	bpx = tile0x * 256;
	bpy = tile0y * 256;
	tms_pxy2latlon((double)bpx/256.0, (double)bpy/256.0, level, &t0clon, &t0clat);
	lldeg2rnd(t0clon, t0clat, &t0cx, &t0cy);

	t0cx += tms->offs.x;
	t0cy += tms->offs.y;

	rnd_trace("TMS draw top-left: %f;%f level=%d\n", lon, lat, level);

	tcx = t0cx; tcy = t0cy;
	tilex = tile0x; tiley = tile0y;

	for(n = 0; n < 256; n++) { /* limit just in case something goes wierd */
		rnd_pixmap_t *pm;
		double size = tile_meter_per_pix[level]*100.0 * 256.0; /* 4.777 is the meter per pixel at this zoom level */
		rnd_coord_t sizec = ceil(size + region->coord_per_pix*0.5); /* round size so there's no gap between tiles */
		tms_tileaddr_t tileaddr;

		tms_tile2key(key, tilex, tiley, level);

		tileaddr.x = tilex;
		tileaddr.y = tiley;
		tileaddr.z = level;
		pm = tms_cache_get(dsg, tms, key, &tileaddr, now);

/*		rnd_trace("   t0c: %ld;%ld size=%ld region0: %ld;%ld %ld;%ld\n", t0cx, t0cy, size, region->view.X1, region->view.Y1, region->view.X2, region->view.Y2);*/
		if (pm != NULL)
			rnd_render->draw_pixmap(rnd_render, tcx, tcy, sizec, sizec, pm);

		tilex++;
		tcx = floor(tcx + size);
		if (tcx - size/2.0 > region->view.X2) {
			tiley++;
			tcy = floor(tcy - size);
			if (tcy + size/2.0 < region->view.Y1)
				break; /* reached end-of-screen */
			tilex = tile0x;
			tcx = t0cx;
		}
	}
}

static int tms_init(rnd_design_t *hl, tms_t *tms, const char *name, const char *url)
{
	TODO("get a callback on layer destroy and uninit");

	tms->maxlev = 19; /* safe default */
	tms->url_segs = tms_url_parse(tms, url);
	if (tms->url_segs == NULL) {
		rnd_message(RND_MSG_ERROR, "TMS: failed to parse TMS url '%s'\n", url);
		return -1;
	}

	tms->name = rnd_strdup(name);
	tms->url = rnd_strdup(url);

	tms_cache_init(hl, tms);
	return 0;
}

static void img_tms_layer_set_offs(osm_layer_t *ly, rnd_coord_t ox, rnd_coord_t oy, int temporary)
{
	tms_t *tms = ly->plugin_data;
	tms->offs.x = ox;
	tms->offs.y = oy;
}

static void img_tms_layer_get_offs(osm_layer_t *ly, rnd_coord_t *ox, rnd_coord_t *oy)
{
	tms_t *tms = ly->plugin_data;
	*ox = tms->offs.x;
	*oy = tms->offs.y;
}

/*** Entry points ***/

const char osm_acts_TMS[] = "TMS(name, tms_url)";
const char osm_acth_TMS[] = "Create the TMS map image layer; name is also used in the changeset source/attribution tag.\n";
fgw_error_t osm_act_TMS(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly;
	const char *url, *name;

	RND_ACT_CONVARG(1, FGW_STR, TMS, name = argv[1].val.str);
	RND_ACT_CONVARG(1, FGW_STR, TMS, url = argv[2].val.str);

	ly = osm_layer_by_name_bottom(osm, name, 1);
	ly->plugin_data = calloc(sizeof(tms_t), 1);
	if (tms_init(hl, ly->plugin_data, name, url) != 0) {
		free(ly->plugin_data);
		return -1;
	}

	ly->plugin_draw = img_tms_layer_draw;
	ly->plugin_set_offs = img_tms_layer_set_offs;
	ly->plugin_get_offs = img_tms_layer_get_offs;
	ly->editable = 0;
	return 0;
}

static void tms_list_iu_ev(rnd_design_t *hidlib, void *user_data, int argc, rnd_event_arg_t argv[])
{
	osm_design_t *osm = (osm_design_t *)hidlib;
	tms_t *tms = NULL;
	long n;


	if ((argc != 3) || (argv[1].type != RND_EVARG_PTR) || (argv[2].type != RND_EVARG_PTR)) {
		rnd_message(RND_MSG_ERROR, "tms_list_iu_ev(): invalid event arguments\n");
		return;
	}

	for(n = 0; n < osm->layers.used; n++) {
		osm_layer_t *ly = osm->layers.array[n];
		if (ly->plugin_draw == img_tms_layer_draw) {
			tms = ly->plugin_data;
			break;
		}
	}

	if (tms != NULL) {
		gds_t *iu = argv[1].d.p, *ioffs = argv[2].d.p;
		if (iu->used > 0)
			gds_append(iu, ';');
		gds_append_str(iu, tms->name);
		if (ioffs->used == 0)
			rnd_append_printf(ioffs, "%.3rm;%.3rm", tms->offs.x, tms->offs.y);
	}
}



static rnd_action_t img_tms_action_list[] = {
	{"tms", osm_act_TMS, osm_acth_TMS, osm_acts_TMS}
};

int pplg_check_ver_img_tms(int ver_needed) { return 0; }

void pplg_uninit_img_tms(void)
{
	rnd_remove_actions_by_cookie(img_tms_cookie);
	rnd_event_unbind_allcookie(img_tms_cookie);
	rnd_conf_plug_unreg("plugins/img_tms/", img_tms_conf_internal, img_tms_cookie);
}

int pplg_init_img_tms(void)
{
	rnd_conf_plug_reg(conf_tms, img_tms_conf_internal, img_tms_cookie);

#define conf_reg(field,isarray,type_name,cpath,cname,desc,flags) \
	rnd_conf_reg_field(conf_tms, field,isarray,type_name,cpath,cname,desc,flags);
#include "img_tms_conf_fields.h"

	RND_REGISTER_ACTIONS(img_tms_action_list, img_tms_cookie);
	rnd_event_bind(OSM_EVENT_LIST_IMAGERY_USED, tms_list_iu_ev, NULL, img_tms_cookie);

	return 0;
}
