/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - server communication
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <math.h>
#include <ctype.h>
#include <librnd/config.h>
#include <librnd/core/hidlib.h>
#include <librnd/core/conf_multi.h>
#include <librnd/core/actions.h>
#include <librnd/core/conf.h>
#include <librnd/core/event.h>
#include <librnd/core/box.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
#include <librnd/hid/pixmap.h>
#include <librnd/core/rnd_conf.h>
#include <librnd/core/compat_fs.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/paths.h>
#include <librnd/core/funchash_core.h>

#include "src_plugins/lib_wgetosm/lib_wgetosm.h"

#include "data.h"
#include "drc.h"
#include "event.h"
#include "style.h"
#include "projection.h"
#include "conf_core.h"
#include "search.h"
#include "history.h"
#include "funchash_core.h"

#include "src_plugins/io_osm_xml/io_osm_xml.h"
#include "src_plugins/dialogs/dlg_tags.h"


#include "net_osm_xml_conf.h"

conf_net_osm_xml_t conf_net_osm_xml;

static const char net_osm_xml_cookie[] = "net_osm";

#include "conf_internal.c"

static void view_get_bbox(osm_design_t *osm, double *lat_bl, double *lon_bl, double *lat_tr, double *lon_tr)
{
	rnd_box_t view;

	rnd_gui->view_get(rnd_gui, &view);

	TODO("TODO#fy5: this lis a librnd 4.1.x workaround for y returned wrong in flip_y=1 mode");
	view.Y2 = osm->hidlib.dwg.Y2 - view.Y2;
	view.Y1 = osm->hidlib.dwg.Y2 - view.Y1;
	rnd_swap(rnd_coord_t, view.Y2, view.Y1);

	rnd_trace("get: %ld;%ld ~ %ld;%ld ch=%ld;%ld\n", view.X1, view.Y1, view.X2, view.Y2, osm->crosshair_x, osm->crosshair_y);

	rnd2lldeg(lon_bl, lat_bl, view.X1, view.Y1);
	rnd2lldeg(lon_tr, lat_tr, view.X2, view.Y2);

	rnd_trace("     %.4f;%.4f ~ %.4f;%.4f\n", *lon_bl, *lat_bl, *lon_tr, *lat_tr);
}

static int update_db(osm_design_t *osm)
{
	double lat_bl, lon_bl, lat_tr, lon_tr; /* bottom-left and top-right lat/lon in deg */
	gds_t url = {0};
	char *tmpfn;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	int lres;

	if (conf_core.editor.offline) {
		rnd_message(RND_MSG_ERROR, "Can't update in offline-mode\n");
		return -1;
	}

	if ((ly != NULL) && !ly->editable) {
		/* e.g. Bing or gpx layer selected; don't load into that, rather create a new layer */
		ly = NULL;
	}

	if ((ly != NULL) && (!ly->editable)) {
		rnd_message(RND_MSG_ERROR, "Can not load osm data onto non-editable (non-osm) layer\n");
		return -1;
	}

	view_get_bbox(osm, &lat_bl, &lon_bl, &lat_tr, &lon_tr);

	gds_append_str(&url, "https://");
	gds_append_str(&url, conf_net_osm_xml.plugins.net_osm_xml.hostname);
	gds_append_str(&url, "/api/0.6/map?bbox=");
	rnd_append_printf(&url, "%.7f,%.7f,%.7f,%.7f\n", lon_bl, lat_bl, lon_tr, lat_tr);

	rnd_trace("     url='%s'\n", url.array);

	tmpfn = rnd_tempfile_name_new("osm_get.xml");
	osm_wget_disk(url.array, tmpfn, 0, NULL);

	rnd_trace("    tmpfn=%s\n", tmpfn);

	rnd_event(&osm->hidlib, RND_EVENT_LOAD_PRE, "s", "osm_get.xml");
	lres = osm_oxml_load_into(osm, tmpfn, &ly);
	rnd_event(&osm->hidlib, RND_EVENT_LOAD_POST, "si", "osm_get.xml", lres);

	osm_data_bbox(osm);
	rnd_box_bump_box(&osm->hidlib.dwg, (rnd_box_t *)&osm->bbox);

	/* if there is no layer selected, we have been loading a new layer; select it */
	if (osm->lysel < 0) {
		int idx;
		for(idx = 0; idx < osm->layers.used; idx++) {
			if (osm->layers.array[idx] == ly) {
				osm_layer_select(osm, idx);
				break;
			}
		}
	}

	rnd_tempfile_unlink(tmpfn);
	gds_uninit(&url);
	return 0;
}

static int update_gpx(osm_design_t *osm)
{
	osm_layer_t *ly = osm_layer_by_name(osm, "gpx", 1);
	double lat_bl, lon_bl, lat_tr, lon_tr; /* bottom-left and top-right lat/lon in deg */
	int lres;
	gds_t url = {0};
	char *tmpfn;

	if (conf_core.editor.offline) {
		rnd_message(RND_MSG_ERROR, "Can't update gpx in offline-mode\n");
		return -1;
	}

/* /api/0.6/trackpoints?bbox=left,bottom,right,top&page=pageNumber*/

	view_get_bbox(osm, &lat_bl, &lon_bl, &lat_tr, &lon_tr);

	gds_append_str(&url, "https://");
	gds_append_str(&url, conf_net_osm_xml.plugins.net_osm_xml.hostname);
	gds_append_str(&url, "/api/0.6/trackpoints?bbox=");
	rnd_append_printf(&url, "%.7f,%.7f,%.7f,%.7f\n", lon_bl, lat_bl, lon_tr, lat_tr);

	rnd_trace("     url='%s'\n", url.array);

	tmpfn = rnd_tempfile_name_new("osm_gpx.xml");
	osm_wget_disk(url.array, tmpfn, 0, NULL);

	rnd_trace("    tmpfn=%s\n", tmpfn);

	rnd_event(&osm->hidlib, RND_EVENT_LOAD_PRE, "s", "osm_gpx.xml");
	lres = osm_oxml_load_into(osm, tmpfn, &ly);
	rnd_event(&osm->hidlib, RND_EVENT_LOAD_POST, "si", "osm_gpx.xml", lres);

	osm_data_bbox(osm);
	rnd_box_bump_box(&osm->hidlib.dwg, (rnd_box_t *)&osm->bbox);

	rnd_tempfile_unlink(tmpfn);
	gds_uninit(&url);
	return 0;
}

const char osm_acts_Update[] = "Update([db|gpx])";
const char osm_acth_Update[] = "Download vector data from the OSM server for editing\n";
fgw_error_t osm_act_Update(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	const char *op = "db";
	rnd_box_t vbx;
	int rv = -1;

	RND_ACT_MAY_CONVARG(1, FGW_STR, Update, op = argv[1].val.str);

	rnd_gui->view_get(rnd_gui, &vbx);

	TODO("TODO#fy5: this lis a librnd 4.1.x workaround for y returned wrong in flip_y=1 mode");
	vbx.Y2 = osm->hidlib.dwg.Y2 - vbx.Y2;
	vbx.Y1 = osm->hidlib.dwg.Y2 - vbx.Y1;
	rnd_swap(rnd_coord_t, vbx.Y2, vbx.Y1);

	if (strcmp(op, "db") == 0)
		rv = update_db(osm);
	else if (strcmp(op, "gpx") == 0)
		rv = update_gpx(osm);
	else {
		rnd_message(RND_MSG_ERROR, "Update(): invalid first argument '%s'\n", op);
		RND_ACT_FAIL(Update);
	}

	rnd_gui->zoom_win(rnd_gui, vbx.X1, vbx.Y1, vbx.X2, vbx.Y2, 0);

	return rv;
}

#include "history_get.c"

const char osm_acts_History[] = "History(object|selected)\nHistory(save|load, path)";
const char osm_acth_History[] = "Download history of an object or of multiple objects\n";
fgw_error_t osm_act_History(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	osm_any_obj_t *obj;
	const char *fn = NULL;
	vtp0_t work = {0};
	int op;

	RND_ACT_CONVARG(1, FGW_KEYWORD, History, op = fgw_keyword(&argv[1]));
	RND_ACT_MAY_CONVARG(2, FGW_STR, History, fn = argv[2].val.str);

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer");
		return -1;
	}


	if (!osm_hist_inited) {
		osm_history_init(&osm_hist);
		osm_hist_inited = 1;
	}


	switch(op) {
		case F_Object:
			obj = osm_search_obj_xy(ly, osm->crosshair_x, osm->crosshair_y);
			if (obj == NULL)
				return -1;
			vtp0_append(&work, obj);
			break;

		case F_Selected:
			osm_search_all_selected_mask(&work, ly, OSM_MASK_WAY);
			break;

		case F_Save:
			if (fn == NULL) {
				rnd_message(RND_MSG_ERROR, "Need a file name");
				return -1;
			}

			osm_oxml_save_history(osm, &osm_hist, fn);

			RND_ACT_IRES(0);
			return 0;

		case F_Load:
			if (fn == NULL) {
				rnd_message(RND_MSG_ERROR, "Need a file name");
				return -1;
			}

			osm_oxml_load_history(&osm_hist, fn);

			RND_ACT_IRES(0);
			return 0;
	}

	if (work.used != 0) {
		history_get(&osm_hist, &work);
		vtp0_uninit(&work);
	}

	RND_ACT_IRES(0);
	return 0;
}

#include "osmchange.c"
#include "diffresult.c"

static osm_id_t osm_ci_changeset_create(osm_tags_t *tags)
{
	osm_wget_opts_t opts = {0};
	gds_t url = {0};
	FILE *f;
	char *freeme, line[1024], *end, *wargs;
	osm_id_t chid = -1;
	char *fn_xml = "/tmp/commit1.xml";
	htsp_entry_t *e;

	TODO("^generate a real temp file name");

	wargs = rnd_concat(
		" --auth-no-challenge"                     /* force sending basic auth */
		" --method=PUT --body-file=", fn_xml,
		NULL);

	/* write the commit xml with tags */
	f = rnd_fopen(NULL, fn_xml, "w");
	fprintf(f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	fprintf(f, "<osm>\n");
	fprintf(f, " <changeset>\n");
	for(e = htsp_first(tags); e != NULL; e = htsp_next(tags, e)) {
		TODO("xml quote strings");
		fprintf(f, "  <tag k=\"%s\" v=\"%s\"/>\n", e->key, (char *)e->value);
	}
	fprintf(f, " </changeset>\n");
	fprintf(f, "</osm>\n");
	fclose(f);

	/* make the http request */
	opts.conf_file = freeme = rnd_concat(rnd_conf.rc.path.home, "/" DOT_OSM_RND "/", conf_net_osm_xml.plugins.net_osm_xml.auth_file_name, NULL);
	opts.extra_args = wargs;

	gds_append_str(&url, "https://");
	gds_append_str(&url, conf_net_osm_xml.plugins.net_osm_xml.hostname);
	gds_append_str(&url, "/api/0.6/changeset/create");

	f = osm_wget_popen(url.array, 0, &opts);
	if (f != NULL) {
		if ((fgets(line, sizeof(line), f)) != NULL) {
			chid = osm_strtoid(line, &end, 10);
rnd_trace("changeset id: '%s' %ld\n", line, (long)chid);
			if ((*end != '\0') && (!isspace(*end)))
				chid = -1;
		}
		fclose(f);
	}

	gds_uninit(&url);
	free(freeme);
	free(wargs);
	return chid;
}

static int osm_ci_changeset_upload(const char *fn_diff, char *fn_diffres, osm_id_t changeset_id, osm_wget_response_t *resp)
{
	osm_wget_opts_t opts = {0};
	gds_t url = {0};
	char *freeme;
	int res;

	opts.conf_file = freeme = rnd_concat(rnd_conf.rc.path.home, "/" DOT_OSM_RND "/", conf_net_osm_xml.plugins.net_osm_xml.auth_file_name, NULL);
	opts.extra_args =
	" --auth-no-challenge";                    /* force sending basic auth */
	opts.post_file = fn_diff;


	gds_append_str(&url, "https://");
	gds_append_str(&url, conf_net_osm_xml.plugins.net_osm_xml.hostname);
	rnd_append_printf(&url, "/api/0.6/changeset/"OSM_PID"/upload", changeset_id);

	res = osm_wget_disk_resp(url.array, fn_diffres, 0, &opts, resp);

	gds_uninit(&url);
	free(freeme);
	return res;
}

static int osm_ci_changeset_close(osm_id_t changeset_id)
{
	osm_wget_opts_t opts = {0};
	gds_t url = {0};
	char *freeme;
	int res;

	opts.conf_file = freeme = rnd_concat(rnd_conf.rc.path.home, "/" DOT_OSM_RND "/", conf_net_osm_xml.plugins.net_osm_xml.auth_file_name, NULL);
	opts.extra_args =
		" --auth-no-challenge"                     /* force sending basic auth */
		" --method=PUT";

	gds_append_str(&url, "https://");
	gds_append_str(&url, conf_net_osm_xml.plugins.net_osm_xml.hostname);
	rnd_append_printf(&url, "/api/0.6/changeset/"OSM_PID"/close", changeset_id);

	res = osm_wget_disk(url.array, "/dev/null", 0, &opts);

	gds_uninit(&url);
	free(freeme);
	return res;
}

static int osm_locdiff_to_log_(osm_any_obj_t *obj, int force_del)
{
	if (!force_del && !obj_changed(obj))
		return 0;

	rnd_message(RND_MSG_INFO, "  - %s #" OSM_PID, osm_obj_name[obj->hdr.type], obj->hdr.id);

	if ((obj->hdr.style != NULL) && !obj->hdr.style->fallback)
		rnd_message(RND_MSG_INFO, " %s", obj->hdr.style->name);

	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:
			rnd_message(RND_MSG_INFO, "\n");
			break;
		case OSM_OBJ_WAY:
			rnd_message(RND_MSG_INFO, " of %ld nodes\n", obj->way.curr->nd.used);
			break;
		case OSM_OBJ_RELATION:
			rnd_message(RND_MSG_INFO, " of %ld nodes\n", obj->relation.curr->members.used);
			break;
		case OSM_OBJ_CHANGESET: case OSM_OBJ_invalid: case OSM_OBJ_vmax:
			rnd_message(RND_MSG_INFO, "<invalid object>\n");
	}

	return 1;
}

static void osm_locdiff_run_drc(osm_layer_t *ly, osm_any_obj_t *obj)
{
	osm_design_t *osm = ly->parent;

	if (!obj_changed(obj))
		return;

	rnd_event(&osm->hidlib, OSM_EVENT_DRC_OBJECT, "p", obj);
}

static vtL0_t my_drc_list;

static int osm_locdiff_to_log(osm_layer_t *ly, osm_rtree_box_t *area, int raise_log, int run_drc)
{
	osm_rtree_it_t it;
	void *o;
	long n;
	int res = 0;
	vtL0_t *old_drc_list;

	if (run_drc) {
		old_drc_list = drc_list;
		drc_list = &my_drc_list;
		osm_drc_clear_list(&my_drc_list);
	}

	rnd_message(RND_MSG_INFO, "--- local diff diff ---\n");

	rnd_message(RND_MSG_INFO, " 1. Object add or edit\n");

	if (area == NULL) {
		for(o = osm_rtree_all_first(&it, &ly->objs); o != NULL; o = osm_rtree_all_next(&it)) {
			res |= osm_locdiff_to_log_(o, 0);
			if (run_drc)
				osm_locdiff_run_drc(ly, o);
		}
	}
	else {
		for(o = osm_rtree_first(&it, &ly->objs, area); o != NULL; o = osm_rtree_next(&it)) {
			res |= osm_locdiff_to_log_(o, 0);
			if (run_drc)
				osm_locdiff_run_drc(ly, o);
		}
	}

	rnd_message(RND_MSG_INFO, " 2. Object removal\n");
	for(n = 0; n < ly->remove_list.used; n++) {
		osm_any_obj_t *obj = ly->remove_list.array[n];
		if ((obj != NULL) && (obj->hdr.id > 0)) {
			osm_locdiff_to_log_(obj, 1);
			res = 1;
		}
	}

	rnd_message(RND_MSG_INFO, "--- end of diff ---\n");

	if (raise_log)
		rnd_actionva(&ly->parent->hidlib, "logdialog", NULL);

	if (run_drc) {
		drc_list = old_drc_list;
		if (my_drc_list.used > 0)
			rnd_event(&ly->parent->hidlib, OSM_EVENT_DRC_PRESENT, "pp", ly, &my_drc_list);
	}


	return res;
}

const char osm_acts_Commit[] = "Commit([force], [survey[=url]], [comment=text])";
const char osm_acth_Commit[] = "Create a changeset from local changes and upload it to the OSM server. If force is specified, go on with the commit even if there are drc errors.\n";
fgw_error_t osm_act_Commit(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	char *fn_diffup, *fn_diffres;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	osm_wget_response_t resp;
	FILE *f;
	osm_id_t changeset_id;
	long num_objs;
	unsigned long orig_cnt;
	int r, succ = 0, force = 0, survey = 0, n;
	osm_tags_t ci_tags = {0};
	const char *survey_url = NULL, *comment = NULL;

	for(n = 1; n < argc; n++) {
		const char *how;
		RND_ACT_CONVARG(n, FGW_STR, Commit, how = argv[n].val.str);
		if (strcmp(how, "force") == 0) {
			force = 1;
		}
		else if (strncmp(how, "survey", 6) == 0) {
			survey = 1;
			if (how[6] == '=')
				survey_url = how+7;
		}
		else if (strncmp(how, "comment=", 8) == 0) {
			comment = how+8;
		}
		else {
			rnd_message(RND_MSG_ERROR, "Commit(): invalid argument '%s'\n", how);
			return -1;
		}
	}

	if (conf_core.editor.offline) {
		rnd_message(RND_MSG_ERROR, "Can't commit in offline-mode\n");
		return -1;
	}

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No layer to commit\n");
		return -1;
	}

	if (!ly->editable) {
		rnd_message(RND_MSG_ERROR, "Can not commit osm data from non-editable (non-osm) layer\n");
		return -1;
	}

	orig_cnt = osm_drc_cnt;
	if (osm_locdiff_to_log(ly, NULL, 1, 1) == 0) {
		rnd_message(RND_MSG_ERROR, "There is no change to commit\n");
		return -1;
	}

	if (!force && (orig_cnt != osm_drc_cnt)) {
		rnd_message(RND_MSG_ERROR, "There are DRC errors, commit cancelled\n");
		return -1;
	}


	osm_tags_init(&ci_tags);
	osm_tag_set(&ci_tags, "created_by", "osm-rnd " OSM_VERS " osm_act_Commit()");

	{ /* build imagery_used, e.g. "Bing;foo" and offset for the first */
		gds_t iu = {0}, ioffs = {0};

		rnd_event(hl, OSM_EVENT_LIST_IMAGERY_USED, "pp", &iu, &ioffs);

		if (iu.used > 0) {
			osm_tag_set(&ci_tags, "imagery_used", iu.array);
			if (strchr(iu.array, ';') == NULL)
				osm_tag_set(&ci_tags, "imagery_offset", ioffs.array);
		}

		gds_uninit(&iu);
		gds_uninit(&ioffs);
	}

	if (survey) {
		char *freeme = NULL;

		osm_tag_set(&ci_tags, "source", "survey");

		if (survey_url == NULL)
			survey_url = freeme = rnd_hid_prompt_for(&osm->hidlib, "Survey ref url", "", "Commit survey source_ref:url");
		if ((survey_url != NULL) && (*survey_url != '\0'))
			osm_tag_set(&ci_tags, "source_ref:url", survey_url);
		free(freeme);
	}

	if (comment != NULL)
		osm_tag_set(&ci_tags, "comment", comment);

	if (osm_rnd_dlg_tags(ly, OSM_OBJ_invalid, 0, &ci_tags) != 0) {
		rnd_message(RND_MSG_INFO, "commit cancelled.\n");
		osm_tags_uninit(&ci_tags);
		return -1;
	}

	changeset_id = osm_ci_changeset_create(&ci_tags);
	if (changeset_id <= 0) {
		rnd_message(RND_MSG_ERROR, "Failed to acquire a new changeset ID.\n");
		osm_tags_uninit(&ci_tags);
		return 0;
	}

rnd_trace("changeset_id=%ld\n", changeset_id);

	TODO("generate temp file name");
	fn_diffup = "A.osm.diff";

	f = rnd_fopen(hl, fn_diffup, "w");
	num_objs = osmchange_gen_xml(f, ly, NULL, changeset_id);
	fclose(f);

	if (num_objs < 0) {
		osm_tags_uninit(&ci_tags);
		return -1;
	}

	TODO("this should be checked earlier so no changeset is created");
	if (num_objs == 0) {
		rnd_message(RND_MSG_ERROR, "There's no change on layer, nothing to commit\n");
		osm_tags_uninit(&ci_tags);
		return 0;
	}

	TODO("generate a temp file here");
	fn_diffres = "UPL.xml";

	r = osm_ci_changeset_upload(fn_diffup, fn_diffres, changeset_id, &resp);
	if ((r != 0) || (resp.status != 200)) {
		rnd_message(RND_MSG_ERROR, "Failed to upload changeset\nStatus: %d", resp.status);
		if (resp.error != NULL)
			rnd_message(RND_MSG_ERROR, " Error: '%s'", resp.error);
		rnd_message(RND_MSG_ERROR, "\n");
	}
	else
		succ = 1;

	osm_wget_response_uninit(&resp);

	if (succ) {
		if (osm_diffresult_apply(ly, fn_diffres) != 0) {
			rnd_message(RND_MSG_ERROR, "**************************************\n");
			rnd_message(RND_MSG_ERROR, "Unexpected error on upload; data in memory broken.\nPlease quit osm-rnd immediately\n");
			rnd_message(RND_MSG_ERROR, "**************************************\n");
		}
		uundo_list_clear(&ly->undo);
	}

	TODO("delete tmp files: fn_diffup, fn_diffres");

	r = osm_ci_changeset_close(changeset_id);
	if (r != 0)
		rnd_message(RND_MSG_ERROR, "Failed to close changeset\n");

	osm_tags_uninit(&ci_tags);
	return r;
}

const char osm_acts_Changeset[] = "Changeset([filename])";
const char osm_acth_Changeset[] = "Save a changeset file in OsmChange format (useful for debugging).\n";
fgw_error_t osm_act_Changeset(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	const char *diffname = "Changeset";
	long num_objs, changeset_id = 1234;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	FILE *f;

	RND_ACT_MAY_CONVARG(1, FGW_STR, Changeset, diffname = argv[1].val.str);

	f = rnd_fopen(hl, diffname, "w");
	num_objs = osmchange_gen_xml(f, ly, NULL, changeset_id);
	fclose(f);

	rnd_message(RND_MSG_INFO, "Saved changeset with %ld objects to '%s'\n", num_objs, diffname);

	return 0;
}

const char osm_acts_Locdiff[] = "Locdiff([filename])";
const char osm_acth_Locdiff[] = "Save a changeset file in OsmChange format (useful for debugging).\n";
fgw_error_t osm_act_Locdiff(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);

	osm_locdiff_to_log(ly, NULL, 1, 1);

	return 0;
}



static rnd_action_t net_osm_xml_action_list[] = {
	{"update", osm_act_Update, osm_acth_Update, osm_acts_Update},
	{"up",     osm_act_Update, osm_acth_Update, osm_acts_Update},
	{"commit", osm_act_Commit, osm_acth_Commit, osm_acts_Commit},
	{"ci",     osm_act_Commit, osm_acth_Commit, osm_acts_Commit},
	{"changeset", osm_act_Changeset, osm_acth_Changeset, osm_acts_Changeset},
	{"locdiff", osm_act_Locdiff, osm_acth_Locdiff, osm_acts_Locdiff},
	{"history", osm_act_History, osm_acth_History, osm_acts_History}
};


int pplg_check_ver_net_osm_xml(int ver_needed) { return 0; }

void pplg_uninit_net_osm_xml(void)
{
	osm_drc_clear_list(&my_drc_list);
	rnd_remove_actions_by_cookie(net_osm_xml_cookie);
	rnd_conf_plug_unreg("plugins/net_osm_xml/", net_osm_xml_conf_internal, net_osm_xml_cookie);
}

int pplg_init_net_osm_xml(void)
{
	rnd_conf_plug_reg(conf_net_osm_xml, net_osm_xml_conf_internal, net_osm_xml_cookie);

#define conf_reg(field,isarray,type_name,cpath,cname,desc,flags) \
	rnd_conf_reg_field(conf_net_osm_xml, field,isarray,type_name,cpath,cname,desc,flags);
#include "net_osm_xml_conf_fields.h"

	RND_REGISTER_ACTIONS(net_osm_xml_action_list, net_osm_xml_cookie);
	return 0;
}
