/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Presets for tagging - field edit dialog
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <ctype.h>

static char *presets_field_dlg_combo(presets_ctx_t *ctx, rnd_hid_row_t *row, lht_node_t *field, lht_node_t *field_key, const char **opts)
{
	lht_node_t *n, *options;
	rnd_hid_dad_buttons_t clbtn[] = {{"Cancel", -1}, {"ok", 0}, {NULL, 0}};
	vtp0_t vect = {0};
	int rv, wopts, i, defidx = -1;
	long used;
	char *res = NULL;
	const char **o;
	RND_DAD_DECL(dlg);

	if (opts == NULL) {
		options = lht_dom_hash_get(field, "options");
		if ((options == NULL) || (options->type != LHT_LIST)) {
			rnd_message(RND_MSG_ERROR, "Presets database error: combo '%s' has no options\n", field->name);
			return NULL;
		}

		for(n = options->data.list.first, i = 0; n != NULL; n = n->next) {
			if (n->type != LHT_TEXT) continue;
			vtp0_append(&vect, n->data.text.value);
			if (strcmp(n->data.text.value, row->cell[1]) == 0)
				defidx = i;
			i++;
		}
		opts = (const char **)vect.array;
		used = vect.used;
	}
	else {
		for(o = opts, i = 0; *o != NULL; o++, i++)
			if (strcmp(*o, row->cell[1]) == 0)
				defidx = i;
	}

	RND_DAD_BEGIN_VBOX(dlg);
		RND_DAD_COMPFLAG(dlg, RND_HATF_EXPFILL);
		RND_DAD_LABEL(dlg, field->name);
		RND_DAD_ENUM(dlg, opts);
			RND_DAD_DEFAULT_NUM(dlg, defidx);
			wopts = RND_DAD_CURRENT(dlg);
		RND_DAD_BUTTON_CLOSES(dlg, clbtn);
	RND_DAD_END(dlg);

	RND_DAD_NEW("preset_field_combo", dlg, field->name, NULL, rnd_true, NULL);
	rv = RND_DAD_RUN(dlg);

	if (rv == 0) {
		i = dlg[wopts].val.lng;
		if ((i >= 0) && (i < used))
			res = rnd_strdup(opts[i]);
	}

	RND_DAD_FREE(dlg);

	vtp0_uninit(&vect);

	return res;
}


static char *presets_field_dlg_text(presets_ctx_t *ctx, rnd_hid_row_t *row, lht_node_t *field, lht_node_t *field_key, const char *extra)
{
	char *val;
	gds_t tmp = {0};

	gds_append_str(&tmp, field->name);
	if(extra != NULL) {
		gds_append(&tmp, ' ');
		gds_append(&tmp, '[');
		gds_append_str(&tmp, extra);
		gds_append(&tmp, ']');
	}

	val = rnd_hid_prompt_for(&osm.hidlib, tmp.array, row->cell[1], "Presets new text value");

	gds_uninit(&tmp);

	return val;
}

RND_INLINE int broken_phone(const char *val)
{
	const char *s;
	int last_dig = 0;

	if (*val != '+')
		return 1;

	for(s = val+1; *s != '\0'; s++) {
		int dig = isdigit(*s);
		if ((*s != '-') && (*s != ' ') && !dig)
			return 1;
		if (!dig && !last_dig) /* duplicate separators */
			return 1;
		last_dig = dig;
	}

	return 0;
}


RND_INLINE int broken_speed(const char *val)
{
	char *end;

	strtol(val, &end, 10);
	while(isspace(*end)) end++;

	if (*end == '\0')
		return 0;

	if (strcmp(end, "mph") == 0)
		return 0;

	return 1; /* unknown/invalid suffix */
}

RND_INLINE int broken_height(const char *val)
{
	char *end;

	strtol(val, &end, 10);

	/* metric? */
	while(isspace(*end)) end++;
	if (*end == '\0')
		return 0;

	/* imperial! require feet */
	if (*end != '\'')
		return 1;
	end++;

	while(isspace(*end)) end++;
	if (*end == '\0')
		return 0;

	/* optional: check for inches */
	strtol(end, &end, 10);
	while(isspace(*end)) end++;
	if (*end != '"')
		return 1;

	end++;
	while(isspace(*end)) end++;
	if (*end == '\0')
		return 0;

	/* there is anything else at the end */
	return 1;
}

RND_INLINE int broken_int(const char *val)
{
	char *end;

	strtol(val, &end, 10);
	while(isspace(*end)) end++;

	if (*end == '\0')
		return 0;

	return 1;
}

RND_INLINE int broken_email(const char *val)
{
	if (strchr(val, '@') == NULL)
		return 1;

	if (strchr(val, ' ') != NULL)
		return 1;

	return 0;
}

RND_INLINE int broken_date(const char *val)
{
	static const char *map="dddd-dd-dd";
	const char *s, *m;

	for(s = val, m = map; *s != '\0'; s++, m++) {
		switch(*m) {
			case '\0':
				return 1; /* reached \0 in m but not in s */
			case 'd':
				if (!isdigit(*s))
					return 1;
				break;
			case '-':
				if (*s != '-')
					return 1;
				break;
		}
	}

	return 0;
}

static void presets_field_dlg(presets_ctx_t *ctx, rnd_hid_row_t *row, lht_node_t *field_key)
{
	lht_node_t *ftype;
	presets_presets_field_type_keys_t itype;
	char *val = NULL;
	lht_node_t *field;

	field = field_key->parent;
	if (field->type == LHT_LIST)
		field = field->parent;

	if (field->type != LHT_HASH) {
		rnd_message(RND_MSG_ERROR, "Presets database error: key '%s' has wrong parent type\n", field_key->name);
		return;
	}

	ftype = lht_dom_hash_get(field, "type");
	if ((ftype == NULL) || (ftype->type != LHT_TEXT)) {
		rnd_message(RND_MSG_ERROR, "Presets database error: field '%s' has invalid or missing type\n", field->name);
		return;
	}


	itype = presets_presets_sphash(ftype->data.text.value);
	switch(itype) {
		/* combo */
		case presets_presets_field_type_combo:
		case presets_presets_field_type_radio:
		case presets_presets_field_type_structureRadio:
			val = presets_field_dlg_combo(ctx, row, field, field_key, NULL);
			break;

		case presets_presets_field_type_check:
			{
				static const char *opts[] = {"yes", "no", NULL};
				val = presets_field_dlg_combo(ctx, row, field, field_key, opts);
			}
			break;


		/* text */
		case presets_presets_field_type_date:
			val = presets_field_dlg_text(ctx, row, field, field_key, "date in YYYY-MM-DD format");
			if ((val != NULL) && broken_date(val))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: invalid date, should be in YYYY-MM-DD\n");
			break;

		case presets_presets_field_type_email:
			val = presets_field_dlg_text(ctx, row, field, field_key, "email@address");
			if ((val != NULL) && broken_email(val))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: invalid email address\n");
			break;

		case presets_presets_field_type_number:
			val = presets_field_dlg_text(ctx, row, field, field_key, "ingeger");
			if ((val != NULL) && broken_int(val))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: number shall be an integer value in meters with no suffix or this for imperial: int' int\"\n");
			break;

		case presets_presets_field_type_roadheight:
			val = presets_field_dlg_text(ctx, row, field, field_key, "height in meters or x'y\"");
			if ((val != NULL) && broken_height(val))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: road height shall be an integer value in meters with no suffix or this for imperial: int' int\"\n");
			break;

		case presets_presets_field_type_roadspeed:
			val = presets_field_dlg_text(ctx, row, field, field_key, "speed; suffix: mph or implicit km/h if omitted");
			if ((val != NULL) && broken_speed(val))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: road speed shall be an integer value with no suffix or 'mph' as suffix\n");
			break;

		case presets_presets_field_type_tel:
			val = presets_field_dlg_text(ctx, row, field, field_key, "phone number: +country-area-local");
			if ((val != NULL) && broken_phone(val))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: phone numbers shall be in +country-area-local format\n");
			break;

		case presets_presets_field_type_url:
			val = presets_field_dlg_text(ctx, row, field, field_key, "url");
			if ((val != NULL) && (strncmp(val, "http://", 7) != 0) && (strncmp(val, "https://", 7) != 0) && (strncmp(val, "ftp://", 6) != 0))
				rnd_message(RND_MSG_ERROR, "Presets sanity check: '%s' doesn't look like a valid url\n", val);
			break;

		/* text with no check */
		case presets_presets_field_type_identifier:
		case presets_presets_field_type_text:
		case presets_presets_field_type_wikidata:
		case presets_presets_field_type_wikipedia:
		case presets_presets_field_type_colour:
			val = presets_field_dlg_text(ctx, row, field, field_key, NULL);
			break;



		case presets_presets_field_type_restriction:
		case presets_presets_field_type_restrictions:
		case presets_presets_field_type_defaultCheck:
		case presets_presets_field_type_manyCombo:
		case presets_presets_field_type_multiCombo:
		case presets_presets_field_type_networkCombo:
		case presets_presets_field_type_typeCombo: /* combo box doing a semicolon list */
		case presets_presets_field_type_semiCombo: /* combo box doing a semicolon list */
		case presets_presets_field_type_directionalCombo: /* pair of combo boxes for the two directions (fwd and backwd) */
		case presets_presets_field_type_localized: /* text with hu: or en: prefix */
		case presets_presets_field_type_textarea:
		case presets_presets_field_type_access:  /* different access types logic for ways */
		case presets_presets_field_type_address:
		case presets_presets_field_type_onewayCheck:
		case presets_presets_field_type_SPHASH_INVALID:
			rnd_message(RND_MSG_ERROR, "Field type %s not yet implemented\n", ftype->data.text.value);
			break;


	}

	if (val != NULL) {
		rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wtags];

		rnd_dad_tree_modify_cell(attr, row, 1, val);
		/* do not free val: ownerhsip passed to the cell */
	}

}
