/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - standard tools
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "crosshair.h"
#include "conf_core.h"
#include "data.h"
#include <librnd/core/compat_misc.h>
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>


typedef struct {
	osm_layer_t *ly;
	rnd_coord_t press_x, press_y, orig_ox, orig_oy;
	int click;
} tool_img_align_t;

static tool_img_align_t img_align;

static void tool_img_align_init(void)
{
	img_align.click = 0;
}

static void tool_img_align_uninit(void)
{
	img_align.click = 0;
}

static void tool_img_align_press_at(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y, osm_any_obj_t *obj_in, int ret_arrow)
{
	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "Select a layer first!\n");
		return;
	}

	if ((ly->plugin_set_offs == NULL) || (ly->plugin_get_offs == NULL)) {
		rnd_message(RND_MSG_ERROR, "Can't align this layer\n");
		if (ret_arrow)
			rnd_tool_select_by_name(&ly->parent->hidlib, "arrow");
		return;
	}

	img_align.ly = ly;
	img_align.press_x = x;
	img_align.press_y = y;
	ly->plugin_get_offs(ly, &img_align.orig_ox, &img_align.orig_oy);
	img_align.click = 1;
}

static void set_offs(rnd_design_t *hl, int temporary)
{
	osm_design_t *osm = (osm_design_t *)hl;
	rnd_coord_t dx = osm->crosshair_x - img_align.press_x + img_align.orig_ox;
	rnd_coord_t dy = osm->crosshair_y - img_align.press_y + img_align.orig_oy;
	img_align.ly->plugin_set_offs(img_align.ly, dx, dy, temporary);
}

static void tool_img_align_press(rnd_design_t *hl)
{
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);

	tool_img_align_press_at(ly, osm->crosshair_x, osm->crosshair_y, NULL, 0);
}

static void tool_img_align_release(rnd_design_t *hl)
{
	if (img_align.click) {
		set_offs(hl, 0);
		img_align.click = 0;
	}
}

static void tool_img_align_adjust_attached_objects(rnd_design_t *hl)
{
	if (img_align.click)
		set_offs(hl, 1);
}

static void tool_img_align_draw_attached(rnd_design_t *hl)
{
	if (img_align.click) {
	}
}

/* XPM */
static const char *img_align_icon[] = {
/* columns rows colors chars-per-pixel */
"21 21 3 1",
"  c #000000",
". c #6EA5D7",
"o c None",
/* pixels */
"ooooooooooooooooooooo",
"oooo   oooooooooooooo",
"oooo . oooooooooooooo",
"oooo . ooooooo   oooo",
"o    .         . oooo",
"o .......      . oooo",
"o    .   oo    .    o",
"oooo .     o....... o",
"oooo .         .    o",
"oooo   ooooooo . oooo",
"oooooooooooooo . oooo",
"oooooooooooooo   oooo",
"ooooooooooooooooooooo",
"ooo  ooo oooo oo   oo",
"ooo  ooo oooo o ooo o",
"oo oo oo oooo o ooooo",
"oo oo oo oooo o ooooo",
"oo    oo oooo o oo  o",
"o oooo o oooo o ooo o",
"o oooo o oooo o ooo o",
"o oooo o   oo oo   oo"
};


static rnd_tool_t osm_rnd_tool_img_align = {
	"img_align", NULL, std_tools_cookie, 50, img_align_icon, RND_TOOL_CURSOR_NAMED("left_ptr"), 0,
	tool_img_align_init,
	tool_img_align_uninit,
	tool_img_align_press,
	tool_img_align_release,
	tool_img_align_adjust_attached_objects,
	tool_img_align_draw_attached,
	NULL, /* undo */
	NULL, /* redo */
	NULL, /* escape */
};

