/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - standard tools
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "tool_mline.h"

#include "crosshair.h"
#include "data.h"
#include "projection.h"
#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
/*#include <librnd/core/misc_util.h>*/
#include <librnd/core/error.h>

osm_tool_mline_t osm_tool_mline;

static void create_mline(rnd_design_t *hl)
{
	double lon1, lat1, lon2, lat2, latm, lonm, tmp, dist, dx, dy;

	rnd2lldeg(&lon1, &lat1, osm_tool_mline.x1, osm_tool_mline.y1);
	rnd2lldeg(&lon2, &lat2, osm_tool_mline.x2, osm_tool_mline.y2);

	/* haversine */
	latm = sin(lat2-lat1)/2;
	lonm = sin(lon2-lon1)/2;
	tmp = latm*latm + cos(lat1)*cos(lat2)*lonm*lonm;

	dx = osm_tool_mline.x1 - osm_tool_mline.x2;
	dy = osm_tool_mline.y1 - osm_tool_mline.y2;

	if ((tmp <= 0) || ((dx == 0) && (dy == 0)))
		return;

	dist = 2 * EARTH_R * asin(sqrt(tmp));


	rnd_message(RND_MSG_INFO, "Distance: haversin: %.3f m | 2d: %.3f m\n", dist / 50.0, sqrt(dx*dx + dy*dy)/100.0);
}

static void tool_mline_init(void)
{
}

static void tool_mline_uninit(void)
{
	osm_tool_mline.clicked = 0;
}

static void tool_mline_press(rnd_design_t *hl)
{
}

static void tool_mline_release(rnd_design_t *hl)
{
	switch(osm_tool_mline.clicked) {
		case 0:
			osm_tool_mline.x1 = osm_tool_mline.x2 = osm.crosshair_x;
			osm_tool_mline.y1 = osm_tool_mline.y2 = osm.crosshair_y;
			osm_tool_mline.clicked = 1;
			break;
		case 1:
			osm_tool_mline.x2 = osm.crosshair_x;
			osm_tool_mline.y2 = osm.crosshair_y;
			create_mline(hl);
			osm_tool_mline.clicked = 0;
			break;
	}
}

static void tool_mline_adjust_attached_objects(rnd_design_t *hl)
{
	if (osm_tool_mline.clicked == 1) {
		osm_tool_mline.x2 = osm.crosshair_x;
		osm_tool_mline.y2 = osm.crosshair_y;
	}
}


static void tool_mline_draw_attached(rnd_design_t *hl)
{
	rnd_hid_set_line_cap(osm_crosshair_gc, rnd_cap_round);
	rnd_hid_set_line_width(osm_crosshair_gc, -1);

	switch(osm_tool_mline.clicked) {
		case 0: break;
		case 1:
			rnd_render->draw_line(osm_crosshair_gc, osm_tool_mline.x1, osm_tool_mline.y1, osm_tool_mline.x2, osm_tool_mline.y2);
			break;
	}
}

rnd_bool tool_mline_undo_act(rnd_design_t *hl)
{
	return 0;
}

rnd_bool tool_mline_redo_act(rnd_design_t *hl)
{
	return 0;
}



/* XPM */
static const char *mline_icon[] = {
/* columns rows colors chars-per-pixel */
"21 21 4 1",
"  c #000000",
". c #7A8584",
"X c #6EA5D7",
"O c None",
/* pixels */
"OOOO.OOOOOOOOOOOOOOOO",
"OOOO.OOOOOOOOOOOOOOOO",
"OOO.OOOOOOOOOOOOOOOOO",
"OOO.XXOOOOOOOOOOOOOOO",
"OOO.OOXXXOOOOOOOOO.OO",
"OO.OOOOOOXXXOOOOOO.OO",
"OO.OOOOOOOOOXXXOO.OOO",
"OOOOOOOOOOOOOOOXX.OOO",
"OOOOOOOOOOOOOOOOO.OOO",
"OOOOOOOOOOOOOOOO.OOOO",
"OOOOOOOOOOOOOOOO.OOOO",
"OOOOOOOOOOOOOOOOOOOOO",
"O OOO O OOO   O OOO O",
"O  O  O OOOO OO  OO O",
"O O O O OOOO OO  OO O",
"O OOO O OOOO OO O O O",
"O OOO O OOOO OO O O O",
"O OOO O OOOO OO OO  O",
"O OOO O OOOO OO OO  O",
"O OOO O   O   O OOO O",
"OOOOOOOOOOOOOOOOOOOOO"
};


static rnd_tool_t osm_rnd_tool_mline = {
	"mline", NULL, std_tools_cookie, 100, mline_icon, RND_TOOL_CURSOR_NAMED("pencil"), 0,
	tool_mline_init,
	tool_mline_uninit,
	tool_mline_press,
	tool_mline_release,
	tool_mline_adjust_attached_objects,
	tool_mline_draw_attached,
	tool_mline_undo_act,
	tool_mline_redo_act
};
